from __future__ import annotations

from functools import reduce
from itertools import chain as flatten
from math import prod
from operator import add
from typing import Sequence, Tuple

import pyqtorch as pyq
import sympy
from pyqtorch.apply import apply_operator
from pyqtorch.matrices import _dagger
from pyqtorch.utils import is_diag
from torch import (
    Tensor,
    argsort,
    bmm,
    cdouble,
    diag_embed,
    diagonal,
    exp,
    linalg,
    ones_like,
    permute,
    tensor,
    transpose,
)
from torch import device as torch_device
from torch.nn import Module

from qadence.backends.utils import (
    finitediff,
    pyqify,
    unpyqify,
)
from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    CompositeBlock,
    MatrixBlock,
    ParametricBlock,
    PrimitiveBlock,
    ScaleBlock,
    TimeEvolutionBlock,
)
from qadence.blocks.block_to_tensor import (
    _block_to_tensor_embedded,
    block_to_diagonal,
    block_to_tensor,
)
from qadence.blocks.primitive import ProjectorBlock
from qadence.operations import (
    U,
    multi_qubit_gateset,
    non_unitary_gateset,
    single_qubit_gateset,
    three_qubit_gateset,
    two_qubit_gateset,
)
from qadence.types import OpName
from qadence.utils import infer_batchsize

from .config import Configuration

# Tdagger is not supported currently
supported_gates = list(set(OpName.list()) - set([OpName.TDAGGER]))
"""The set of supported gates.

Tdagger is currently not supported.
"""


def is_single_qubit_chain(block: AbstractBlock) -> bool:
    return (
        isinstance(block, (ChainBlock))
        and block.n_supports == 1
        and all([isinstance(b, (ParametricBlock, PrimitiveBlock)) for b in block])
        and not any([isinstance(b, (ScaleBlock, U)) for b in block])
    )


def convert_observable(
    block: AbstractBlock, n_qubits: int, config: Configuration = None
) -> Sequence[Module]:
    return [PyQObservable(block, n_qubits, config)]


def convert_block(
    block: AbstractBlock, n_qubits: int = None, config: Configuration = None
) -> Sequence[Module]:
    qubit_support = block.qubit_support
    if n_qubits is None:
        n_qubits = max(qubit_support) + 1

    if config is None:
        config = Configuration()

    if isinstance(block, ScaleBlock):
        return [ScalePyQOperation(n_qubits, block, config)]

    elif isinstance(block, AddBlock):
        ops = list(flatten(*(convert_block(b, n_qubits, config) for b in block.blocks)))
        return [AddPyQOperation(n_qubits, ops)]

    elif isinstance(block, TimeEvolutionBlock):
        return [
            PyQHamiltonianEvolution(
                qubit_support=qubit_support,
                n_qubits=n_qubits,
                block=block,
                config=config,
            )
        ]
    elif isinstance(block, MatrixBlock):
        return [PyQMatrixBlock(block, n_qubits, config)]
    elif isinstance(block, CompositeBlock):
        ops = list(flatten(*(convert_block(b, n_qubits, config) for b in block.blocks)))
        if is_single_qubit_chain(block) and config.use_single_qubit_composition:
            return [PyQComposedBlock(ops, qubit_support, n_qubits, config)]
        else:
            # NOTE: without wrapping in a pyq.QuantumCircuit here the kron/chain
            # blocks won't be properly nested which leads to incorrect results from
            # the `AddBlock`s. For example:
            # add(chain(Z(0), Z(1))) has to result in the following (pseudo-code)
            # AddPyQOperation(pyq.QuantumCircuit(Z, Z))
            # as opposed to
            # AddPyQOperation(Z, Z)
            # which would be wrong.
            return [pyq.QuantumCircuit(n_qubits, ops)]
    elif isinstance(block, tuple(non_unitary_gateset)):
        if isinstance(block, ProjectorBlock):
            projector = getattr(pyq, block.name)
            if block.name == OpName.N:
                return [projector(target=qubit_support)]
            else:
                return [projector(qubit_support=qubit_support, ket=block.ket, bra=block.bra)]
        else:
            return [getattr(pyq, block.name)(qubit_support[0])]
    elif isinstance(block, tuple(single_qubit_gateset)):
        pyq_cls = getattr(pyq, block.name)
        if isinstance(block, ParametricBlock):
            if isinstance(block, U):
                op = pyq_cls(qubit_support[0], *config.get_param_name(block))
            else:
                op = pyq_cls(qubit_support[0], config.get_param_name(block)[0])
        else:
            op = pyq_cls(qubit_support[0])
        return [op]
    elif isinstance(block, tuple(two_qubit_gateset)):
        pyq_cls = getattr(pyq, block.name)
        if isinstance(block, ParametricBlock):
            op = pyq_cls(qubit_support[0], qubit_support[1], config.get_param_name(block)[0])
        else:
            op = pyq_cls(qubit_support[0], qubit_support[1])
        return [op]
    elif isinstance(block, tuple(three_qubit_gateset) + tuple(multi_qubit_gateset)):
        block_name = block.name[1:] if block.name.startswith("M") else block.name
        pyq_cls = getattr(pyq, block_name)
        if isinstance(block, ParametricBlock):
            op = pyq_cls(qubit_support[:-1], qubit_support[-1], config.get_param_name(block)[0])
        else:
            op = pyq_cls(qubit_support[:-1], qubit_support[-1])
        return [op]
    else:
        raise NotImplementedError(
            f"Non supported operation of type {type(block)}. "
            "In case you are trying to run an `AnalogBlock`, make sure you "
            "specify the `device_specs` in your `Register` first."
        )


class PyQMatrixBlock(Module):
    def __init__(self, block: MatrixBlock, n_qubits: int, config: Configuration = None):
        super().__init__()
        self.n_qubits = n_qubits
        self.qubits = block.qubit_support
        self.register_buffer("mat", block.matrix.unsqueeze(2))
        self.mat: Tensor
        self._device: torch_device = self.mat.device

    def forward(self, state: Tensor, _: dict[str, Tensor] = None) -> Tensor:
        return apply_operator(state, self.mat, self.qubits, self.n_qubits)

    @property
    def device(self) -> torch_device:
        return self._device

    def to(self, device: torch_device) -> PyQMatrixBlock:
        self.mat = self.mat.to(device)
        self._device = device
        return self


class PyQComposedBlock(pyq.QuantumCircuit):
    def __init__(
        self,
        ops: list[Module],
        qubits: Tuple[int, ...],
        n_qubits: int,
        config: Configuration = None,
    ):
        """Compose a chain of single qubit operations on the same qubit into a single.

        call to _apply_batch_gate.
        """
        super().__init__(n_qubits, ops)
        self.qubits = qubits

    def forward(self, state: Tensor, values: dict[str, Tensor] | None = None) -> Tensor:
        batch_size = infer_batchsize(values)
        return apply_operator(
            state, self.unitary(values, batch_size), self.qubits, self.n_qubits, batch_size
        )

    def unitary(self, values: dict[str, Tensor] | None, batch_size: int) -> Tensor:
        batch_first_perm = (2, 0, 1)
        undo_perm = tuple(argsort(tensor(batch_first_perm)))

        def _expand(m: Tensor) -> Tensor:
            if len(m.size()) == 2:
                m = m.unsqueeze(2).repeat(
                    1, 1, batch_size
                )  # Primitive gates are 2D, so we expand them.
            elif m.shape != (2, 2, batch_size):
                m = m.repeat(1, 1, batch_size)  # In case a tensor is 3D doesnt have batch_size.
            return m

        def _batch_first(m: Tensor) -> Tensor:
            return permute(m, batch_first_perm)  # This returns shape (batch_size, 2, 2)

        def _batch_last(m: Tensor) -> Tensor:
            return permute(
                m, undo_perm
            )  # We need to undo the permute since PyQ expects (2, 2, batch_size).

        # We reverse the list of tensors here since matmul is not commutative.

        return _batch_last(
            reduce(
                bmm,
                (_batch_first(_expand(op.unitary(values))) for op in reversed(self.operations)),
            )
        )


class PyQObservable(Module):
    def __init__(self, block: AbstractBlock, n_qubits: int, config: Configuration = None):
        super().__init__()
        if config is None:
            config = Configuration()
        self.n_qubits = n_qubits
        if block._is_diag_pauli and not block.is_parametric:
            self.register_buffer("operation", block_to_diagonal(block, tuple(range(n_qubits))))
            self._forward = lambda self, state, values: pyqify(
                self.operation * unpyqify(state), n_qubits=self.n_qubits
            )
        else:
            self.operation = pyq.QuantumCircuit(
                n_qubits,
                convert_block(block, n_qubits, config),
            )
            self._forward = lambda self, state, values: self.operation(state, values)
        self._device = self.operation.device

    def run(self, state: Tensor, values: dict[str, Tensor]) -> Tensor:
        return self._forward(self, state, values)

    def forward(self, state: Tensor, values: dict[str, Tensor]) -> Tensor:
        return pyq.overlap(state, self.run(state, values))

    @property
    def device(self) -> torch_device:
        return self._device

    def to(self, device: torch_device) -> PyQObservable:
        self.operation = self.operation.to(device)
        self._device = device
        return self


class PyQHamiltonianEvolution(Module):
    def __init__(
        self,
        qubit_support: Tuple[int, ...],
        n_qubits: int,
        block: TimeEvolutionBlock,
        config: Configuration,
    ):
        super().__init__()
        self.qubit_support = qubit_support
        self.n_qubits = n_qubits
        self.param_names = config.get_param_name(block)
        self.block = block
        self.hmat: Tensor

        if isinstance(block.generator, AbstractBlock) and not block.generator.is_parametric:
            hmat = block_to_tensor(
                block.generator,
                qubit_support=self.qubit_support,
                use_full_support=False,
            )
            hmat = hmat.permute(1, 2, 0)
            self.register_buffer("hmat", hmat)
            self._hamiltonian = lambda self, values: self.hmat

        elif isinstance(block.generator, Tensor):
            m = block.generator.to(dtype=cdouble)
            hmat = block_to_tensor(
                MatrixBlock(m, qubit_support=block.qubit_support),
                qubit_support=self.qubit_support,
                use_full_support=False,
            )
            hmat = hmat.permute(1, 2, 0)
            self.register_buffer("hmat", hmat)
            self._hamiltonian = lambda self, values: self.hmat

        elif isinstance(block.generator, sympy.Basic):
            self._hamiltonian = (
                lambda self, values: values[self.param_names[1]].squeeze(3).permute(1, 2, 0)
            )
            # FIXME Why are we squeezing
        else:

            def _hamiltonian(self: PyQHamiltonianEvolution, values: dict[str, Tensor]) -> Tensor:
                hmat = _block_to_tensor_embedded(
                    block.generator,  # type: ignore[arg-type]
                    values=values,
                    qubit_support=self.qubit_support,
                    use_full_support=False,
                    device=self.device,
                )
                return hmat.permute(1, 2, 0)

            self._hamiltonian = _hamiltonian

        self._time_evolution = lambda values: values[self.param_names[0]]
        self._device: torch_device = (
            self.hmat.device if hasattr(self, "hmat") else torch_device("cpu")
        )

    def _unitary(self, hamiltonian: Tensor, time_evolution: Tensor) -> Tensor:
        self.batch_size = max(hamiltonian.size()[2], len(time_evolution))
        diag_check = tensor(
            [is_diag(hamiltonian[..., i]) for i in range(hamiltonian.size()[2])], device=self.device
        )

        def _evolve_diag_operator(hamiltonian: Tensor, time_evolution: Tensor) -> Tensor:
            evol_operator = diagonal(hamiltonian) * (-1j * time_evolution).view((-1, 1))
            evol_operator = diag_embed(exp(evol_operator))
            return transpose(evol_operator, 0, -1)

        def _evolve_matrixexp_operator(hamiltonian: Tensor, time_evolution: Tensor) -> Tensor:
            evol_operator = transpose(hamiltonian, 0, -1) * (-1j * time_evolution).view((-1, 1, 1))
            evol_operator = linalg.matrix_exp(evol_operator)
            return transpose(evol_operator, 0, -1)

        evolve_operator = (
            _evolve_diag_operator if bool(prod(diag_check)) else _evolve_matrixexp_operator
        )
        return evolve_operator(hamiltonian, time_evolution)

    def unitary(self, values: dict[str, Tensor]) -> Tensor:
        """The evolved operator given current parameter values for generator and time evolution."""
        return self._unitary(self._hamiltonian(self, values), self._time_evolution(values))

    def jacobian_time(self, values: dict[str, Tensor]) -> Tensor:
        """Approximate jacobian of the evolved operator with respect to time evolution."""
        return finitediff(
            lambda t: self._unitary(time_evolution=t, hamiltonian=self._hamiltonian(self, values)),
            values[self.param_names[0]],
        )

    def jacobian_generator(self, values: dict[str, Tensor]) -> Tensor:
        """Approximate jacobian of the evolved operator with respect to generator parameter(s)."""
        if len(self.param_names) > 2:
            raise NotImplementedError(
                "jacobian_generator does not support generators\
                                        with more than 1 parameter."
            )

        def _generator(val: Tensor) -> Tensor:
            val_copy = values.copy()
            val_copy[self.param_names[1]] = val
            hmat = _block_to_tensor_embedded(
                self.block.generator,  # type: ignore[arg-type]
                values=val_copy,
                qubit_support=self.qubit_support,
                use_full_support=False,
                device=self.device,
            )
            return hmat.permute(1, 2, 0)

        return finitediff(
            lambda v: self._unitary(
                time_evolution=self._time_evolution(values), hamiltonian=_generator(v)
            ),
            values[self.param_names[1]],
        )

    def dagger(self, values: dict[str, Tensor]) -> Tensor:
        """Dagger of the evolved operator given the current parameter values."""
        return _dagger(self.unitary(values))

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor],
    ) -> Tensor:
        return apply_operator(
            state,
            self.unitary(values),
            self.qubit_support,
            self.n_qubits,
            self.batch_size,
        )

    @property
    def device(self) -> torch_device:
        return self._device

    def to(self, device: torch_device) -> PyQHamiltonianEvolution:
        if hasattr(self, "hmat"):
            self.hmat = self.hmat.to(device)
        self._device = device
        return self


class AddPyQOperation(pyq.QuantumCircuit):
    def __init__(self, n_qubits: int, operations: list[Module]):
        super().__init__(n_qubits=n_qubits, operations=operations)

    def forward(self, state: Tensor, values: dict[str, Tensor]) -> Tensor:
        return reduce(add, (op(state, values) for op in self.operations))


class ScalePyQOperation(pyq.QuantumCircuit):
    def __init__(self, n_qubits: int, block: ScaleBlock, config: Configuration):
        if not isinstance(block.block, PrimitiveBlock):
            raise NotImplementedError(
                "The pyqtorch backend can currently only scale `PrimitiveBlock` types.\
                Please use the following transpile function on your circuit first:\
                from qadence.transpile import scale_primitive_blocks_only"
            )
        ops = convert_block(block.block, n_qubits, config)
        assert len(ops) == 1
        super().__init__(n_qubits, ops)
        (self.param_name,) = config.get_param_name(block)
        self.qubit_support = self.operations[0].qubit_support

    def forward(self, state: Tensor, values: dict[str, Tensor]) -> Tensor:
        return apply_operator(state, self.unitary(values), self.qubit_support, self.n_qubits)

    def unitary(self, values: dict[str, Tensor]) -> Tensor:
        thetas = values[self.param_name]
        return thetas * self.operations[0].unitary(values)

    def dagger(self, values: dict[str, Tensor]) -> Tensor:
        return _dagger(self.unitary(values))

    def jacobian(self, values: dict[str, Tensor]) -> Tensor:
        return values[self.param_name] * ones_like(self.unitary(values))



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Union
from warnings import warn

from sympy import Expr, Heaviside, exp
from torch import Tensor

from qadence.parameters import Parameter, evaluate
from qadence.types import PI

# FIXME: Clarify the roles of these values in the context
# device specification and how they relate with the
# maximum values for delta and omega.
GLOBAL_MAX_AMPLITUDE = 300
GLOBAL_MAX_DETUNING = 2 * PI * 2000
LOCAL_MAX_AMPLITUDE = 3
LOCAL_MAX_DETUNING = 2 * PI * 20

TWeight = Union[str, float, Tensor, Parameter]


def sigmoid(x: Tensor, a: float, b: float) -> Expr:
    return 1.0 / (1.0 + exp(-a * (x + b)))


@dataclass
class AddressingPattern:
    """Semi-local addressing pattern."""

    n_qubits: int
    """Number of qubits in register."""

    weights_amp: dict[int, TWeight]
    """List of weights for fixed amplitude pattern that cannot be changed during the execution."""

    weights_det: dict[int, TWeight]
    """List of weights for fixed detuning pattern that cannot be changed during the execution."""

    amp: TWeight = LOCAL_MAX_AMPLITUDE
    """Maximum amplitude of the amplitude pattern felt by a single qubit."""

    det: TWeight = LOCAL_MAX_DETUNING
    """Maximum detuning of the detuning pattern felt by a single qubit."""

    def _validate_weights(
        self,
        weights: dict[int, TWeight],
    ) -> None:
        for v in weights.values():
            if not isinstance(v, (str, Parameter)):
                if not (v >= 0.0 and v <= 1.0):
                    raise ValueError("Addressing pattern weights must sum fall in range [0.0, 1.0]")

    def _constrain_weights(
        self,
        weights: dict[int, TWeight],
    ) -> dict:
        # augment weight dict if needed
        weights = {
            i: Parameter(0.0)
            if i not in weights
            else (Parameter(weights[i]) if not isinstance(weights[i], Parameter) else weights[i])
            for i in range(self.n_qubits)
        }

        # restrict weights to [0, 1] range - equal to 0 everywhere else
        weights = {
            k: v if v.is_number else abs(v * (sigmoid(v, 20, 1) - sigmoid(v, 20.0, -1)))  # type: ignore [union-attr]
            for k, v in weights.items()
        }

        return weights

    def _constrain_max_vals(self) -> None:
        # enforce constraints:
        # 0 <= amp <= GLOBAL_MAX_AMPLITUDE
        # 0 <= abs(det) <= GLOBAL_MAX_DETUNING
        self.amp = abs(
            self.amp
            * (
                Heaviside(self.amp + GLOBAL_MAX_AMPLITUDE)  # type: ignore [operator]
                - Heaviside(self.amp - GLOBAL_MAX_AMPLITUDE)  # type: ignore [operator]
            )
        )
        self.det = -abs(
            self.det
            * (
                Heaviside(self.det + GLOBAL_MAX_DETUNING)
                - Heaviside(self.det - GLOBAL_MAX_DETUNING)
            )
        )

    def _create_local_constraint(self, val: Expr, weights: dict, max_val: float) -> dict:
        # enforce local constraints:
        # amp * w_amp_i < LOCAL_MAX_AMPLITUDE or
        # abs(det) * w_det_i < LOCAL_MAX_DETUNING
        local_constr = {k: val * v for k, v in weights.items()}
        local_constr = {k: Heaviside(v) - Heaviside(v - max_val) for k, v in local_constr.items()}

        return local_constr

    def _create_global_constraint(self, val: Expr, weights: dict, max_val: float) -> Expr:
        # enforce global constraints:
        # amp * sum(w_amp_0, w_amp_1, ...) < GLOBAL_MAX_AMPLITUDE or
        # abs(det) * sum(w_det_0, w_det_1, ...) < GLOBAL_MAX_DETUNING
        weighted_vals_global = val * sum([v for v in weights.values()])
        weighted_vals_global = Heaviside(weighted_vals_global) - Heaviside(
            weighted_vals_global - max_val
        )

        return weighted_vals_global

    def __post_init__(self) -> None:
        # validate amplitude/detuning weights
        self._validate_weights(self.weights_amp)
        self._validate_weights(self.weights_det)

        # validate maximum global amplitude/detuning values
        if not isinstance(self.amp, (str, Parameter)):
            if self.amp > GLOBAL_MAX_AMPLITUDE:
                warn("Maximum absolute value of amplitude is exceeded")
        elif isinstance(self.amp, str):
            self.amp = Parameter(self.amp, trainable=True)
        if not isinstance(self.det, (str, Parameter)):
            if abs(self.det) > GLOBAL_MAX_DETUNING:
                warn("Maximum absolute value of detuning is exceeded")
        elif isinstance(self.det, str):
            self.det = Parameter(self.det, trainable=True)

        # constrain amplitude/detuning parameterized weights to [0.0, 1.0] interval
        self.weights_amp = self._constrain_weights(self.weights_amp)
        self.weights_det = self._constrain_weights(self.weights_det)

        # constrain max global amplitude and detuning to strict interval
        self._constrain_max_vals()

        # create additional local and global constraints for amplitude/detuning masks
        self.local_constr_amp = self._create_local_constraint(
            self.amp, self.weights_amp, LOCAL_MAX_AMPLITUDE
        )
        self.local_constr_det = self._create_local_constraint(
            -self.det, self.weights_det, LOCAL_MAX_DETUNING
        )
        self.global_constr_amp = self._create_global_constraint(
            self.amp, self.weights_amp, GLOBAL_MAX_AMPLITUDE
        )
        self.global_constr_det = self._create_global_constraint(
            -self.det, self.weights_det, GLOBAL_MAX_DETUNING
        )

        # validate number of qubits in mask
        if max(list(self.weights_amp.keys())) >= self.n_qubits:
            raise ValueError("Amplitude weight specified for non-existing qubit")
        if max(list(self.weights_det.keys())) >= self.n_qubits:
            raise ValueError("Detuning weight specified for non-existing qubit")

    def evaluate(self, weights: dict, values: dict) -> dict:
        # evaluate weight expressions with actual values
        return {k: evaluate(v, values, as_torch=True).flatten() for k, v in weights.items()}  # type: ignore [union-attr]

    def _to_dict(self) -> dict:
        return {field.name: getattr(self, field.name) for field in fields(self)}

    @classmethod
    def _from_dict(cls, d: dict) -> AddressingPattern | None:
        return cls(**d) if len(d) > 0 else None



from __future__ import annotations

from typing import Any

import numpy as np
import torch
from torch import Tensor


def numpy_to_tensor(
    x: np.ndarray,
    device: torch.device = torch.device("cpu"),
    dtype: torch.dtype = torch.float64,
    requires_grad: bool = False,
) -> Tensor:
    """This only copies the numpy array if device or dtype are different than the ones of x."""
    return torch.as_tensor(x, dtype=dtype, device=device).requires_grad_(requires_grad)


def promote_to_tensor(x: Tensor | np.ndarray | float, requires_grad: bool = True) -> Tensor:
    """Convert the given type into a torch.Tensor."""
    if isinstance(x, float):
        return torch.tensor([[x]], requires_grad=requires_grad)
    elif isinstance(x, np.ndarray):
        return numpy_to_tensor(x, requires_grad=requires_grad)
    elif isinstance(x, Tensor):
        return x.requires_grad_(requires_grad)
    else:
        raise ValueError(f"Don't know how to promote {type(x)} to Tensor")


def promote_to(x: Tensor, dtype: Any) -> float | np.ndarray | Tensor:
    if dtype == float:
        assert x.size() == (1, 1)
        return x[0, 0].item()
    elif dtype == np.ndarray:
        return x.detach().cpu().numpy()
    elif dtype == Tensor:
        return x
    else:
        raise ValueError(f"Don't know how to convert Tensor to {dtype}")



from __future__ import annotations

from typing import cast

import numpy as np
import torch
from torch import Tensor

from qadence import BackendName
from qadence.backends.api import backend_factory
from qadence.backends.pulser.backend import Backend
from qadence.blocks import block_to_tensor
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.analog import ConstantAnalogRotation, InteractionBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.noise import Noise
from qadence.operations import AnalogRot
from qadence.transpile import apply_fn_to_blocks
from qadence.utils import Endianness


def zne(noise_levels: Tensor, zne_datasets: list[list]) -> Tensor:
    poly_fits = []
    for dataset in zne_datasets:  # Looping over batched observables.
        batched_observable: list = []
        n_params = len(dataset[0])
        for p in range(n_params):  # Looping over the batched parameters.
            rearranged_dataset = [s[p] for s in dataset]
            # Polynomial fit function.
            poly_fit = np.poly1d(
                np.polyfit(noise_levels, rearranged_dataset, len(noise_levels) - 1)
            )
            # Return the zero-noise extrapolated value.
            batched_observable.append(poly_fit(0.0))
        poly_fits.append(batched_observable)

    return torch.tensor(poly_fits)


def pulse_experiment(
    backend: Backend,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor],
    noise: Noise,
    stretches: Tensor,
    endianness: Endianness,
    state: Tensor | None = None,
) -> Tensor:
    def mutate_params(block: AbstractBlock, stretch: float) -> AbstractBlock:
        """Closure to retrieve and stretch analog parameters."""
        # Check for stretchable analog block.
        if isinstance(block, (ConstantAnalogRotation, InteractionBlock)):
            stretched_duration = block.parameters.duration * stretch
            stretched_omega = block.parameters.omega / stretch
            stretched_delta = block.parameters.delta / stretch
            # The Hamiltonian scaling has no effect on the phase parameter
            phase = block.parameters.phase
            qubit_support = block.qubit_support
            return AnalogRot(
                duration=stretched_duration,
                omega=stretched_omega,
                delta=stretched_delta,
                phase=phase,
                qubit_support=qubit_support,
            )
        return block

    zne_datasets = []
    noisy_density_matrices: list = []
    for stretch in stretches:
        # FIXME: Iterating through the circuit for every stretch
        # and rebuilding the block leaves is inefficient.
        # Best to retrieve the parameters once
        # and rebuild the blocks.
        stre = stretch.item()
        block = apply_fn_to_blocks(circuit.block, mutate_params, stre)
        stretched_register = circuit.register.rescale_coords(stre)
        stretched_circuit = QuantumCircuit(stretched_register, block)
        conv_circuit = backend.circuit(stretched_circuit)
        noisy_density_matrices.append(
            # Contain a single experiment result for the stretch.
            backend.run_dm(
                conv_circuit,
                param_values=param_values,
                state=state,
                noise=noise,
                endianness=endianness,
            )[0]
        )
    # Convert observable to Numpy types compatible with QuTip simulations.
    # Matrices are flipped to match QuTip conventions.
    converted_observable = [np.flip(block_to_tensor(obs).numpy()) for obs in observable]
    # Create ZNE datasets by looping over batches.
    for observable in converted_observable:
        # Get expectation values at the end of the time serie [0,t]
        # at intervals of the sampling rate.
        zne_datasets.append(
            [
                [dm.expect(observable)[0][-1] for dm in density_matrices]
                for density_matrices in noisy_density_matrices
            ]
        )
    # Zero-noise extrapolate.
    extrapolated_exp_values = zne(
        noise_levels=stretches,
        zne_datasets=zne_datasets,
    )
    return extrapolated_exp_values


def noise_level_experiment(
    backend: Backend,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor],
    noise: Noise,
    endianness: Endianness,
    state: Tensor | None = None,
) -> Tensor:
    noise_probs = noise.options.get("noise_probs")
    zne_datasets: list = []
    # Get noisy density matrices.
    conv_circuit = backend.circuit(circuit)
    noisy_density_matrices = backend.run_dm(
        conv_circuit, param_values=param_values, state=state, noise=noise, endianness=endianness
    )
    # Convert observable to Numpy types compatible with QuTip simulations.
    # Matrices are flipped to match QuTip conventions.
    converted_observable = [np.flip(block_to_tensor(obs).numpy()) for obs in observable]
    # Create ZNE datasets by looping over batches.
    for observable in converted_observable:
        # Get expectation values at the end of the time serie [0,t]
        # at intervals of the sampling rate.
        zne_datasets.append(
            [
                [dm.expect(observable)[0][-1] for dm in density_matrices]
                for density_matrices in noisy_density_matrices
            ]
        )
    # Zero-noise extrapolate.
    extrapolated_exp_values = zne(noise_levels=noise_probs, zne_datasets=zne_datasets)
    return extrapolated_exp_values


def analog_zne(
    backend_name: BackendName,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor] = {},
    state: Tensor | None = None,
    measurement: Measurements | None = None,
    noise: Noise | None = None,
    mitigation: Mitigations | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    assert noise
    assert mitigation
    backend = backend_factory(backend=BackendName.PULSER, diff_mode=None)
    backend = cast(Backend, backend)
    noise_model = mitigation.options.get("noise_model", None)
    if noise_model is None:
        KeyError(f"A noise model should be choosen from {Noise.list()}. Got {noise_model}.")
    stretches = mitigation.options.get("stretches", None)
    if stretches is not None:
        extrapolated_exp_values = pulse_experiment(
            backend=backend,
            circuit=circuit,
            observable=observable,
            param_values=param_values,
            noise=noise,
            stretches=stretches,
            endianness=endianness,
            state=state,
        )
    else:
        extrapolated_exp_values = noise_level_experiment(
            backend=backend,
            circuit=circuit,
            observable=observable,
            param_values=param_values,
            noise=noise,
            endianness=endianness,
            state=state,
        )
    return extrapolated_exp_values


def mitigate(
    backend_name: BackendName,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor] = {},
    state: Tensor | None = None,
    measurement: Measurements | None = None,
    noise: Noise | None = None,
    mitigation: Mitigations | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    mitigated_exp = analog_zne(
        backend_name=backend_name,
        circuit=circuit,
        observable=observable,
        param_values=param_values,
        state=state,
        measurement=measurement,
        noise=noise,
        mitigation=mitigation,
        endianness=endianness,
    )
    return mitigated_exp



from __future__ import annotations

from typing import Callable, Union

from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeRemainingColumn
from torch.nn import DataParallel, Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from qadence.logger import get_logger
from qadence.ml_tools.config import TrainConfig
from qadence.ml_tools.data import DictDataLoader, data_to_device
from qadence.ml_tools.optimize_step import optimize_step
from qadence.ml_tools.printing import print_metrics, write_tensorboard
from qadence.ml_tools.saveload import load_checkpoint, write_checkpoint

logger = get_logger(__name__)


def train(
    model: Module,
    dataloader: Union[None, DataLoader, DictDataLoader],
    optimizer: Optimizer,
    config: TrainConfig,
    loss_fn: Callable,
    device: str = "cpu",
    optimize_step: Callable = optimize_step,
    write_tensorboard: Callable = write_tensorboard,
) -> tuple[Module, Optimizer]:
    """Runs the training loop with gradient-based optimizer.

    Assumes that `loss_fn` returns a tuple of (loss,
    metrics: dict), where `metrics` is a dict of scalars. Loss and metrics are
    written to tensorboard. Checkpoints are written every
    `config.checkpoint_every` steps (and after the last training step).  If a
    checkpoint is found at `config.folder` we resume training from there.  The
    tensorboard logs can be viewed via `tensorboard --logdir /path/to/folder`.

    Args:
        model: The model to train.
        dataloader: dataloader of different types. If None, no data is required by
            the model
        optimizer: The optimizer to use.
        config: `TrainConfig` with additional training options.
        loss_fn: Loss function returning (loss: float, metrics: dict[str, float])
        device: String defining device to train on, pass 'cuda' for GPU.
        optimize_step: Customizable optimization callback which is called at every iteration.=
            The function must have the signature `optimize_step(model,
            optimizer, loss_fn, xs, device="cpu")` (see the example below).
            Apart from the default we already supply three other optimization
            functions `optimize_step_evo`, `optimize_step_grad_norm`, and
            `optimize_step_inv_dirichlet`. Learn more about how to use this in
            the [Advancded features](../../tutorials/advanced) tutorial of the
            documentation.
        write_tensorboard: Customizable tensorboard logging callback which is
            called every `config.write_every` iterations. The function must have
            the signature `write_tensorboard(writer, loss, metrics, iteration)`
            (see the example below).

    Example:
    ```python exec="on" source="material-block"
    from pathlib import Path
    import torch
    from itertools import count
    from qadence import Parameter, QuantumCircuit, Z
    from qadence import hamiltonian_factory, hea, feature_map, chain
    from qadence.models import QNN
    from qadence.ml_tools import TrainConfig, train_with_grad, to_dataloader

    n_qubits = 2
    fm = feature_map(n_qubits)
    ansatz = hea(n_qubits=n_qubits, depth=3)
    observable = hamiltonian_factory(n_qubits, detuning = Z)
    circuit = QuantumCircuit(n_qubits, fm, ansatz)

    model = QNN(circuit, observable, backend="pyqtorch", diff_mode="ad")
    batch_size = 1
    input_values = {"phi": torch.rand(batch_size, requires_grad=True)}
    pred = model(input_values)

    ## lets prepare the train routine

    cnt = count()
    criterion = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)

    def loss_fn(model: torch.nn.Module, data: torch.Tensor) -> tuple[torch.Tensor, dict]:
        next(cnt)
        x, y = data[0], data[1]
        out = model(x)
        loss = criterion(out, y)
        return loss, {}

    tmp_path = Path("/tmp")
    n_epochs = 5
    batch_size = 25
    config = TrainConfig(
        folder=tmp_path,
        max_iter=n_epochs,
        checkpoint_every=100,
        write_every=100,
    )
    x = torch.linspace(0, 1, batch_size).reshape(-1, 1)
    y = torch.sin(x)
    data = to_dataloader(x, y, batch_size=batch_size, infinite=True)
    train_with_grad(model, data, optimizer, config, loss_fn=loss_fn)
    ```
    """

    # Move model to device before optimizer is loaded
    if isinstance(model, DataParallel):
        model = model.module.to(device)
    else:
        model = model.to(device)
    # load available checkpoint
    init_iter = 0
    if config.folder:
        model, optimizer, init_iter = load_checkpoint(config.folder, model, optimizer)
        logger.debug(f"Loaded model and optimizer from {config.folder}")
    # initialize tensorboard
    writer = SummaryWriter(config.folder, purge_step=init_iter)

    ## Training
    progress = Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(elapsed_when_finished=True),
    )

    with progress:
        dl_iter = iter(dataloader) if dataloader is not None else None

        # outer epoch loop
        for iteration in progress.track(range(init_iter, init_iter + config.max_iter)):
            try:
                # in case there is not data needed by the model
                # this is the case, for example, of quantum models
                # which do not have classical input data (e.g. chemistry)
                if dataloader is None:
                    loss, metrics = optimize_step(model, optimizer, loss_fn, None)
                    loss = loss.item()

                elif isinstance(dataloader, (DictDataLoader, DataLoader)):
                    data = data_to_device(next(dl_iter), device)  # type: ignore[arg-type]
                    loss, metrics = optimize_step(model, optimizer, loss_fn, data)

                else:
                    raise NotImplementedError(
                        f"Unsupported dataloader type: {type(dataloader)}. "
                        "You can use e.g. `qadence.ml_tools.to_dataloader` to build a dataloader."
                    )

                if iteration % config.print_every == 0 and config.verbose:
                    print_metrics(loss, metrics, iteration)

                if iteration % config.write_every == 0:
                    write_tensorboard(writer, loss, metrics, iteration)

                if config.folder:
                    if iteration % config.checkpoint_every == 0:
                        write_checkpoint(config.folder, model, optimizer, iteration)

            except KeyboardInterrupt:
                print("Terminating training gracefully after the current iteration.")
                break

    # Final writing and checkpointing
    if config.folder:
        write_checkpoint(config.folder, model, optimizer, iteration)
    write_tensorboard(writer, loss, metrics, iteration)
    writer.close()

    return model, optimizer



from __future__ import annotations

import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from graphviz import Graph

from .themes import BaseTheme, Black, Dark, Light, White

THEMES = {"light": Light, "dark": Dark, "black": Black, "white": White}
LAYOUTS = ["LR", "TB"]


class QuantumCircuitDiagram:
    """This class plots a quantum circuit using Graphviz."""

    __valid_layouts = LAYOUTS
    __themes = THEMES

    __default_graph_attr = {
        "compound": "true",  # This helps to draw edges when clusters do not have to show
        "splines": "false",  # This helps edges perpendicular to wires to occur on the same axis
    }

    def __init__(
        self,
        layout: str = "LR",
        nb_wires: int = 1,
        theme: BaseTheme | str = "light",
        measurement: bool = False,
        **kwargs: Any,
    ):
        self._layout = self._validate_layout(layout)
        self.nb_wires = self._validate_nb_wires(nb_wires)
        self.theme = self._validate_theme(theme)
        self.graph = Graph()
        self.graph.attr(**kwargs)
        # each wire is represented by a list of nodes
        self._wires: List[List[Node]] = [[] for _ in range(self.nb_wires)]
        self._clusters: List[Cluster] = []
        self._measurement: bool = measurement
        self._start_nodes: List[Node] = []
        self._end_nodes: List[Node] = []

    @staticmethod
    def _validate_nb_wires(nb_wires: int) -> int:
        if nb_wires < 1:
            raise ValueError(
                f"Invalid nb_wires {nb_wires}. Only positive integers greater "
                f"than zero are supported."
            )
        return nb_wires

    def _get_graph_attr(self) -> Dict:
        graph_attr = {
            "rankdir": self._layout,
        }
        graph_attr.update(self.__default_graph_attr)
        graph_attr.update(self.theme.get_graph_attr())
        return graph_attr

    def _validate_layout(self, layout: str) -> str:
        layout = layout.upper()
        if layout not in self.__valid_layouts:
            raise ValueError(f"Invalid layout {layout}. Supported: {self.__valid_layouts}")
        return layout

    def _validate_theme(self, theme: BaseTheme | str) -> BaseTheme:
        if isinstance(theme, BaseTheme):
            return theme

        theme = theme.lower()
        if theme not in self.__themes:
            raise ValueError(f"Invalid theme {theme}. Supported: {list(self.__themes.keys())}")
        return self.__themes[theme]()

    def _validate_wire(self, wire: int) -> int:
        if wire >= self.nb_wires:
            raise ValueError(
                f"Invalid wire {wire}. This circuit has {self.nb_wires} wires numbered from 0 "
                f"to {self.nb_wires - 1}"
            )
        return wire

    def _validate_wires_couple(self, wire1: int, wire2: int) -> Tuple[int, int]:
        """Helper to ensure we are not creating a swap gate or a control gate on the same wire."""
        wire1 = self._validate_wire(wire1)
        wire2 = self._validate_wire(wire2)
        if wire1 == wire2:
            raise ValueError("Invalid wires couple: they can't be the same.")
        return wire1, wire2

    def _create_start_node(self, wire: int) -> Node:
        node = self.create_node(
            wire,
            append=False,
            shape="plaintext",
            label=f"{wire}",
            color=self.theme.color,
            fontcolor=self.theme.color,
            fontname=self.theme.fontname,
            fontsize=str(float(self.theme.fontsize) * 3 / 4),
        )
        self._start_nodes.append(node)
        node.generate()
        return node

    def _create_end_node(self, wire: int) -> Node:
        kwargs = {"label": "", "shape": "plaintext"}
        if self._measurement:
            kwargs.update({"image": self.theme.load_measurement_icon(), "fixedsize": "true"})
        node = self.create_node(wire, self, append=False, **kwargs)
        self._end_nodes.append(node)
        node.generate()
        return node

    def _create_swap_departure_node(
        self, wire: int, parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None
    ) -> Node:
        pw = self.theme.get_edge_attr()["penwidth"]
        return self.create_node(
            wire,
            self if parent is None else parent,
            shape="octagon",
            width="0.03",
            height="0.03",
            style="filled",
            color=self.theme.color,
            fillcolor=self.theme.color,
            penwidth="1",
            label="",
        )

    def _create_swap_arrival_node(
        self,
        wire: int,
        swap_departure_node: Node = None,
        parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None,
    ) -> Node:
        pw = self.theme.get_edge_attr()["penwidth"]
        return self.create_node(
            wire,
            self if parent is None else parent,
            swap_departure_node=swap_departure_node,
            shape="octagon",
            width="0.03",
            height="0.03",
            style="filled",
            color=self.theme.color,
            fillcolor=self.theme.color,
            penwidth="1",
            label="",
        )

    def _create_control_gate_departure_node(
        self,
        wire: int,
        parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None,
        **kwargs: Any,
    ) -> Node:
        return self.create_node(wire=wire, parent=self if parent is None else parent, **kwargs)

    def _create_control_gate_arrival_node(
        self,
        wire: int,
        control_gate_departure_node: Node,
        parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None,
    ) -> None:
        self.create_node(
            wire=wire,
            parent=self if parent is None else parent,
            control_gate_departure_node=control_gate_departure_node,
            label="",
            shape="point",
            width="0.2",
            color=self.theme.color,
        )

    def _connect(
        self,
        node1: Node,
        node2: Node,
        **kwargs: str,
    ) -> None:
        """It creates an edge between node1 and node2."""
        edge_attr = self.theme.get_edge_attr()
        edge_attr.update(kwargs)
        if (
            not getattr(node1.parent, "show", True) and node1.parent == node2.parent
        ):  # if we are connecting 2 nodes inside a cluster with show: false
            edge_attr.update({"style": "invis"})
        else:
            if not getattr(node1.parent, "show", True):
                edge_attr.update({"ltail": node1.parent.get_name()})
            if not getattr(node2.parent, "show", True):
                edge_attr.update({"lhead": node2.parent.get_name()})
        self.graph.edge(node1.id, node2.id, **edge_attr)

    def create_control_gate(
        self,
        from_wire: int,
        to_wires: List[int],
        parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None,
        **kwargs: Any,
    ) -> None:
        """
        A control gate consists of a node in a wire, and a list of nodes on other different wires.

        All wires are connected with vertical edges between them.
        """
        for to_wire in to_wires:
            from_wire, to_wire = self._validate_wires_couple(from_wire, to_wire)
        departure_node = self._create_control_gate_departure_node(
            from_wire, parent=parent, **kwargs
        )
        for to_wire in to_wires:
            self._create_control_gate_arrival_node(to_wire, departure_node, parent=parent)

    def create_swap_gate(
        self,
        wire_1: int,
        wire_2: int,
        parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None,
    ) -> None:
        """
        A swap gate.

        It consists in 4 invisible nodes, 2 invisible edges on the same wire, and 2 dotted
        edges on the destination wires, empty edges on the others
        """
        wire_1, wire_2 = self._validate_wires_couple(wire_1, wire_2)
        departure_node_1 = self._create_swap_departure_node(wire_1, parent=parent)
        departure_node_2 = self._create_swap_departure_node(wire_2, parent=parent)
        self._create_swap_arrival_node(wire_2, swap_departure_node=departure_node_1, parent=parent)
        self._create_swap_arrival_node(wire_1, swap_departure_node=departure_node_2, parent=parent)

    def create_node(
        self,
        wire: int,
        parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None,
        append: bool = True,
        **kwargs: Union[str, Node, None],
    ) -> Node:
        """Create a node on the specified wire for the Quantum Circuit Diagram or for a cluster."""
        self._validate_wire(wire)
        node = Node(wire, self if parent is None else parent, **kwargs)
        if append:
            self._wires[wire].append(node)
        return node

    def create_identity_node(
        self, wire: int, parent: Optional[Union[QuantumCircuitDiagram, Cluster]] = None
    ) -> "Node":
        pw = self.theme.get_edge_attr()["penwidth"]
        return self.create_node(
            wire,
            self if parent is None else parent,
            shape="square",
            label="",
            width="0",
            height="0",
            style="filled",
            color=self.theme.color,
            penwidth=str(float(pw) * 0.5),
        )

    def create_cluster(
        self,
        label: str,
        parent: Optional[Cluster] = None,
        show: bool = True,
        **kwargs: Any,
    ) -> Cluster:
        """
        A cluster is a tagged sub diagram of the Quantum Circuit Diagram.

        It's implemented using
        Graphviz's sub graphs.
        """
        cluster = Cluster(
            self if parent is None else parent,
            label,
            show=show,
            **kwargs,
        )
        self._clusters.append(cluster)
        return cluster

    def _build(self) -> None:
        self.graph.graph_attr.update(self._get_graph_attr())
        # keep track of control gates to connect them after generation
        control_gates: list[list[Node]] = []
        # keep track of swap gates to connect them after generation
        swap_gates: list[list[Node]] = []
        for i, wire in enumerate(self._wires):
            start_node = self._create_start_node(i)
            prev_node = start_node
            for j, node in enumerate(wire):
                node.generate()
                if getattr(node, "swap_departure_node", None):
                    self._connect(prev_node, node, style="invis")
                else:
                    self._connect(prev_node, node)
                prev_node = node
                if getattr(node, "control_gate_departure_node", None):
                    control_gates.append([node.control_gate_departure_node, node])  # type:ignore
                if getattr(node, "swap_departure_node", None):
                    swap_gates.append([node.swap_departure_node, node])  # type:ignore
            end_node = self._create_end_node(i)
            self._connect(wire[-1], end_node)

        for cluster in reversed(self._clusters):
            cluster.generate()

        # Connect control and swap gates
        for gate in control_gates:
            self._connect(gate[0], gate[1], constraint="false")
        for gate in swap_gates:
            self._connect(gate[0], gate[1])

        # Connect all starting and ending nodes with invisible edges, this keeps wires in order
        def _linelink(nodes: list) -> None:
            for i in range(1, len(nodes)):
                self._connect(nodes[i - 1], nodes[i], constraint="false", style="invis")

        _linelink(self._start_nodes)
        _linelink(self._end_nodes)

    @staticmethod
    def _runtime() -> str:
        try:
            ipy_str = str(type(get_ipython()))  # type: ignore[name-defined] # noqa
            if "zmqshell" in ipy_str:
                return "jupyter"
            elif "terminal" in ipy_str:
                return "ipython"
            else:
                raise
        except Exception as e:
            return "terminal"

    def show(self) -> Graph:
        self._build()
        if not self._runtime() == "jupyter":
            self.graph.view()
        return self.graph

    def savefig(self, filename: str) -> None:
        fn = Path(filename)
        self._build()
        self.graph.format = fn.suffix[1:]
        self.graph.render(str(fn.parent / fn.stem), view=False)
        # self.graph.save(filename)


class Node:
    def __init__(
        self,
        wire: int,
        parent: Any,
        **kwargs: Any,
    ) -> None:
        self.id = uuid.uuid4().hex
        self.wire = wire
        self.parent = parent
        self.control_gate_departure_node = kwargs.pop("control_gate_departure_node", None)
        self.swap_departure_node = kwargs.pop("swap_departure_node", None)
        self.attrs = {"group": str(wire)}
        self.attrs.update(kwargs)

    def generate(self) -> None:
        self.parent.graph.node(self.id, **self.attrs)


class Cluster:
    def __init__(
        self,
        parent: Any,
        label: str,
        show: bool = True,
        **kwargs: str,
    ) -> None:
        self._id: str = uuid.uuid4().hex
        self.parent: Union[QuantumCircuitDiagram, Cluster] = parent
        self._label: str = label
        graph_attr: Dict[str, str] = {"label": self._label}
        graph_attr.update(kwargs)
        self.graph: Graph = Graph(name=self.get_name(), graph_attr=graph_attr)
        self.show: bool = show

    def __repr__(self) -> str:
        return f"{self._id} - {self._label}"

    def _get_qcd(self) -> QuantumCircuitDiagram:
        parent = self.parent
        while isinstance(parent, Cluster):
            parent = parent.parent
        return parent

    @property
    def theme(self) -> BaseTheme:
        return self._get_qcd().theme

    @property
    def nb_wires(self) -> int:
        return self._get_qcd().nb_wires

    def get_name(self) -> str:
        return f"cluster_{self._id}"

    def generate(self) -> None:
        self.parent.graph.subgraph(self.graph)

    def create_node(self, wire: int, **kwargs: str) -> Node:
        if not self.show:
            kwargs.update({"style": "invis"})
        return self._get_qcd().create_node(wire, parent=self, append=True, **kwargs)

    def create_identity_node(self, wire: int) -> None:
        self._get_qcd().create_identity_node(wire, self)

    def create_cluster(self, label: str, show: bool = True, **kwargs: Any) -> Cluster:
        return self._get_qcd().create_cluster(label, parent=self, show=show, **kwargs)

    def create_control_gate(
        self, from_wire: int, to_wires: List[int], label: str, **kwargs: Any
    ) -> None:
        self._get_qcd().create_control_gate(from_wire, to_wires, label=label, parent=self, **kwargs)

    def create_swap_gate(self, wire_1: int, wire_2: int) -> None:
        self._get_qcd().create_swap_gate(wire_1, wire_2, self)



from __future__ import annotations

from jax import config

from .differentiable_backend import DifferentiableBackend
from .differentiable_expectation import DifferentiableExpectation

config.update("jax_enable_x64", True)



from __future__ import annotations

from importlib import import_module

from torch import cdouble, set_default_dtype
from torch import float64 as torchfloat64

from .analog import *
from .backend import *
from .backends import *
from .blocks import *
from .circuit import *
from .constructors import *
from .engines import *
from .exceptions import *
from .execution import *
from .measurements import *
from .ml_tools import *
from .models import *
from .noise import *
from .operations import *
from .overlap import *
from .parameters import *
from .register import *
from .serialization import *
from .states import *
from .transpile import *
from .types import *
from .utils import *

DEFAULT_FLOAT_DTYPE = torchfloat64
DEFAULT_COMPLEX_DTYPE = cdouble
set_default_dtype(DEFAULT_FLOAT_DTYPE)

"""Fetch the functions defined in the __all__ of each sub-module.

Import to the qadence name space. Make sure each added submodule has the respective definition:

    - `__all__ = ["function0", "function1", ...]`

Furthermore, add the submodule to the list below to automatically build
the __all__ of the qadence namespace. Make sure to keep alphabetical ordering.
"""

list_of_submodules = [
    ".analog",
    ".backends",
    ".blocks",
    ".circuit",
    ".constructors",
    ".noise",
    ".exceptions",
    ".execution",
    ".measurements",
    ".ml_tools",
    ".models",
    ".operations",
    ".overlap",
    ".parameters",
    ".register",
    ".serialization",
    ".states",
    ".transpile",
    ".types",
    ".utils",
]

__all__ = []
for submodule in list_of_submodules:
    __all_submodule__ = getattr(import_module(submodule, package="qadence"), "__all__")
    __all__ += __all_submodule__



from __future__ import annotations

import importlib
from dataclasses import dataclass
from typing import Callable, Counter, cast

PROTOCOL_TO_MODULE = {
    "readout": "qadence.noise.readout",
}


@dataclass
class Noise:
    DEPHASING = "dephasing"
    DEPOLARIZING = "depolarizing"
    READOUT = "readout"

    def __init__(self, protocol: str, options: dict = dict()) -> None:
        self.protocol: str = protocol
        self.options: dict = options

    def get_noise_fn(self) -> Callable:
        try:
            module = importlib.import_module(PROTOCOL_TO_MODULE[self.protocol])
        except KeyError:
            ImportError(f"The module corresponding to the protocol {self.protocol} is not found.")
        fn = getattr(module, "add_noise")
        return cast(Callable, fn)

    def _to_dict(self) -> dict:
        return {"protocol": self.protocol, "options": self.options}

    @classmethod
    def _from_dict(cls, d: dict) -> Noise | None:
        if d:
            return cls(d["protocol"], **d["options"])
        return None

    @classmethod
    def list(cls) -> list:
        return list(filter(lambda el: not el.startswith("__"), dir(cls)))


def apply_noise(noise: Noise, samples: list[Counter]) -> list[Counter]:
    """Apply noise to samples."""
    error_fn = noise.get_noise_fn()
    # Get the number of qubits from the sample keys.
    n_qubits = len(list(samples[0].keys())[0])
    # Get the number of shots from the sample values.
    n_shots = sum(samples[0].values())
    noisy_samples: list = error_fn(
        counters=samples, n_qubits=n_qubits, options=noise.options, n_shots=n_shots
    )
    return noisy_samples



from __future__ import annotations

from torch.utils.tensorboard import SummaryWriter


def print_metrics(loss: float | None, metrics: dict, iteration: int) -> None:
    msg = " ".join(
        [f"Iteration {iteration: >7} | Loss: {loss:.7f} -"]
        + [f"{k}: {v.item():.7f}" for k, v in metrics.items()]
    )
    print(msg)


def write_tensorboard(
    writer: SummaryWriter, loss: float | None, metrics: dict, iteration: int
) -> None:
    writer.add_scalar("loss", loss, iteration)
    for key, arg in metrics.items():
        writer.add_scalar(key, arg, iteration)


def log_hyperparams(writer: SummaryWriter, hyperparams: dict, metrics: dict) -> None:
    writer.add_hparams(hyperparams, metrics)



from __future__ import annotations

from typing import Any, Callable

import torch
from torch.nn import Module
from torch.optim import Optimizer


def optimize_step(
    model: Module,
    optimizer: Optimizer,
    loss_fn: Callable,
    xs: dict | list | torch.Tensor | None,
) -> tuple[torch.Tensor | float, dict | None]:
    """Default Torch optimize step with closure.

    This is the default optimization step which should work for most
    of the standard use cases of optimization of Torch models

    Args:
        model (Module): The input model
        optimizer (Optimizer): The chosen Torch optimizer
        loss_fn (Callable): A custom loss function
        xs (dict | list | torch.Tensor | None): the input data. If None it means
            that the given model does not require any input data

    Returns:
        tuple: tuple containing the model, the optimizer, a dictionary with
            the collected metrics and the compute value loss
    """

    loss, metrics = None, {}

    def closure() -> Any:
        # NOTE: We need the nonlocal as we can't return a metric dict and
        # because e.g. LBFGS calls this closure multiple times but for some
        # reason the returned loss is always the first one...
        nonlocal metrics, loss
        optimizer.zero_grad()
        loss, metrics = loss_fn(model, xs)
        loss.backward(retain_graph=True)
        return loss.item()

    optimizer.step(closure)
    # return the loss/metrics that are being mutated inside the closure...
    return loss, metrics



from __future__ import annotations

from collections import Counter
from typing import Any, Callable

import numpy as np
import torch
from torch import Tensor

from qadence.backend import BackendConfiguration
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.utils import chain, kron, tag
from qadence.circuit import QuantumCircuit
from qadence.divergences import js_divergence
from qadence.measurements import Measurements
from qadence.models.quantum_model import QuantumModel
from qadence.operations import SWAP, H, I, S, Z
from qadence.transpile import reassign
from qadence.types import BackendName, DiffMode, OverlapMethod

# Modules to be automatically added to the qadence namespace
__all__ = ["Overlap", "OverlapMethod"]


def _cswap(control: int, target1: int, target2: int) -> AbstractBlock:
    # define projectors on control qubit
    p0 = 0.5 * I(control) + 0.5 * Z(control)
    p1 = 0.5 * I(control) + (-0.5) * Z(control)

    # construct controlled-SWAP block
    cswap_blocks = kron(p0, I(target1), I(target2)) + kron(p1, SWAP(target1, target2))
    cswap = tag(cswap_blocks, f"CSWAP({control}, {target1}, {target2})")

    return cswap


def _controlled_unitary(control: int, unitary_block: AbstractBlock) -> AbstractBlock:
    n_qubits = unitary_block.n_qubits

    # define projectors on control qubit
    p0 = 0.5 * I(control) + 0.5 * Z(control)
    p1 = 0.5 * I(control) + (-0.5) * Z(control)

    # shift qubit support of unitary
    shifted_unitary_block = reassign(unitary_block, {i: control + i + 1 for i in range(n_qubits)})

    # construct controlled-U block
    cu_blocks = kron(p0, *[I(control + i + 1) for i in range(n_qubits)]) + kron(
        p1, shifted_unitary_block
    )
    cu = tag(cu_blocks, f"c-U({control}, {shifted_unitary_block.qubit_support})")

    return cu


def _is_counter_list(lst: list[Counter]) -> bool:
    return all(map(lambda x: isinstance(x, Counter), lst)) and isinstance(lst, list)


def _select_overlap_method(
    method: OverlapMethod,
    backend: BackendName,
    bra_circuit: QuantumCircuit,
    ket_circuit: QuantumCircuit,
) -> tuple[Callable, QuantumCircuit, QuantumCircuit]:
    if method == OverlapMethod.EXACT:
        fn = overlap_exact

        def _overlap_fn(
            param_values: dict,
            bra_calc_fn: Callable,
            bra_state: Tensor | None,
            ket_calc_fn: Callable,
            ket_state: Tensor | None,
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            kets = ket_calc_fn(param_values["ket"], ket_state)
            overlap = fn(bras, kets)
            return overlap

    elif method == OverlapMethod.JENSEN_SHANNON:

        def _overlap_fn(
            param_values: dict,
            bra_calc_fn: Callable,
            bra_state: Tensor | None,
            ket_calc_fn: Callable,
            ket_state: Tensor | None,
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            kets = ket_calc_fn(param_values["ket"], ket_state)
            overlap = overlap_jensen_shannon(bras, kets)
            return overlap

    elif method == OverlapMethod.COMPUTE_UNCOMPUTE:
        # create a single circuit from bra and ket circuits
        bra_circuit = QuantumCircuit(
            bra_circuit.n_qubits, bra_circuit.block, ket_circuit.block.dagger()
        )
        ket_circuit = None  # type: ignore[assignment]

        def _overlap_fn(  # type: ignore [misc]
            param_values: dict, bra_calc_fn: Callable, bra_state: Tensor | None, *_: Any
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            overlap = overlap_compute_uncompute(bras)
            return overlap

    elif method == OverlapMethod.SWAP_TEST:
        if backend == BackendName.BRAKET:
            raise ValueError("SWAP test method is not supported by the Braket backend.")

        n_qubits = bra_circuit.n_qubits

        # shift qubit support of bra and ket circuit blocks
        shifted_bra_block = reassign(bra_circuit.block, {i: i + 1 for i in range(n_qubits)})
        shifted_ket_block = reassign(
            ket_circuit.block, {i: i + n_qubits + 1 for i in range(n_qubits)}
        )
        ket_circuit = None  # type: ignore[assignment]

        # construct swap test circuit
        state_blocks = kron(shifted_bra_block, shifted_ket_block)
        cswap_blocks = chain(*[_cswap(0, n + 1, n + 1 + n_qubits) for n in range(n_qubits)])
        swap_test_blocks = chain(H(0), state_blocks, cswap_blocks, H(0))
        bra_circuit = QuantumCircuit(2 * n_qubits + 1, swap_test_blocks)

        def _overlap_fn(  # type: ignore [misc]
            param_values: dict, bra_calc_fn: Callable, bra_state: Tensor | None, *_: Any
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            overlap = overlap_swap_test(bras)
            return overlap

    elif method == OverlapMethod.HADAMARD_TEST:
        if backend == BackendName.BRAKET:
            raise ValueError("Hadamard test method is not supported by the Braket backend.")

        n_qubits = bra_circuit.n_qubits

        # construct controlled bra and ket blocks
        c_bra_block = _controlled_unitary(0, bra_circuit.block)
        c_ket_block = _controlled_unitary(0, ket_circuit.block.dagger())

        # construct swap test circuit for Re part
        re_blocks = chain(H(0), c_bra_block, c_ket_block, H(0))
        bra_circuit = QuantumCircuit(n_qubits + 1, re_blocks)

        # construct swap test circuit for Im part
        im_blocks = chain(H(0), c_bra_block, c_ket_block, S(0), H(0))
        ket_circuit = QuantumCircuit(n_qubits + 1, im_blocks)

        def _overlap_fn(
            param_values: dict,
            bra_calc_fn: Callable,
            bra_state: Tensor | None,
            ket_calc_fn: Callable,
            ket_state: Tensor | None,
        ) -> Tensor:
            bras = bra_calc_fn(param_values["bra"], bra_state)
            kets = ket_calc_fn(param_values["ket"], ket_state)
            overlap = overlap_hadamard_test(bras, kets)
            return overlap

    return _overlap_fn, bra_circuit, ket_circuit


def overlap_exact(bras: Tensor, kets: Tensor) -> Tensor:
    """Calculate overlap using exact quantum mechanical definition.

    Args:
        bras (Tensor): full bra wavefunctions
        kets (Tensor): full ket wavefunctions

    Returns:
        Tensor: overlap tensor containing values of overlap of each bra with each ket
    """
    return torch.abs(torch.sum(bras.conj() * kets, dim=1)) ** 2


def fidelity(bras: Tensor, kets: Tensor) -> Tensor:
    return overlap_exact(bras, kets)


def overlap_jensen_shannon(bras: list[Counter], kets: list[Counter]) -> Tensor:
    """Calculate overlap from bitstring counts using Jensen-Shannon divergence method.

    Args:
        bras (list[Counter]): bitstring counts corresponding to bra wavefunctions
        kets (list[Counter]): bitstring counts corresponding to ket wavefunctions

    Returns:
        Tensor: overlap tensor containing values of overlap of each bra with each ket
    """
    return 1 - torch.tensor([js_divergence(p, q) for p, q in zip(bras, kets)])


def overlap_compute_uncompute(bras: Tensor | list[Counter]) -> Tensor:
    """Calculate overlap using compute-uncompute method.

    From full wavefunctions or bitstring counts.

    Args:
        bras (Tensor | list[Counter]): full bra wavefunctions or bitstring counts

    Returns:
        Tensor: overlap tensor containing values of overlap of each bra with zeros ket
    """
    if isinstance(bras, Tensor):
        # calculate exact overlap of full bra wavefunctions with |0> state
        overlap = torch.abs(bras[:, 0]) ** 2

    elif isinstance(bras, list):
        # estimate overlap as the fraction of shots when "0..00" bitstring was observed
        n_qubits = len(list(bras[0].keys())[0])
        n_shots = sum(list(bras[0].values()))
        overlap = torch.tensor([p["0" * n_qubits] / n_shots for p in bras])

    return overlap


def overlap_swap_test(bras: Tensor | list[Counter]) -> Tensor:
    """Calculate overlap using swap test method.

    From full wavefunctions or bitstring counts.

    Args:
        bras (Tensor | list[Counter]): full bra wavefunctions or bitstring counts

    Returns:
        Tensor: overlap tensor
    """
    if isinstance(bras, Tensor):
        n_qubits = int(np.log2(bras.shape[1]))

        # define measurement operator  |0><0| x I
        proj_op = torch.tensor([[1.0, 0.0], [0.0, 0.0]])
        ident_op = torch.diag(torch.tensor([1.0 for _ in range(2 ** (n_qubits - 1))]))
        meas_op = torch.kron(proj_op, ident_op).type(torch.complex128)

        # estimate overlap from ancilla qubit measurement
        prob0 = (bras.conj() * torch.matmul(meas_op, bras.t()).t()).sum(dim=1).real

    elif _is_counter_list(bras):
        # estimate overlap as the fraction of shots when 0 was observed on ancilla qubit
        n_qubits = len(list(bras[0].keys())[0])
        n_shots = sum(list(bras[0].values()))
        prob0 = torch.tensor(
            [
                sum(map(lambda k, v: v if k[0] == "0" else 0, p.keys(), p.values())) / n_shots
                for p in bras
            ]
        )
    else:
        raise TypeError("Incorrect type passed for bras argument.")

    # construct final overlap tensor
    overlap = 2 * prob0 - 1

    return overlap


def overlap_hadamard_test(
    bras_re: Tensor | list[Counter], bras_im: Tensor | list[Counter]
) -> Tensor:
    """Calculate overlap using Hadamard test method.

    From full wavefunctions or bitstring counts.

    Args:
        bras_re (Tensor | list[Counter]): full bra wavefunctions or bitstring counts
        for estimation of overlap's real part
        bras_im (Tensor | list[Counter]): full bra wavefunctions or bitstring counts
        for estimation of overlap's imaginary part

    Returns:
        Tensor: overlap tensor
    """
    if isinstance(bras_re, Tensor) and isinstance(bras_im, Tensor):
        n_qubits = int(np.log2(bras_re.shape[1]))

        # define measurement operator  |0><0| x I
        proj_op = torch.tensor([[1.0, 0.0], [0.0, 0.0]])
        ident_op = torch.diag(torch.tensor([1.0 for _ in range(2 ** (n_qubits - 1))]))
        meas_op = torch.kron(proj_op, ident_op).type(torch.complex128)

        # estimate overlap from ancilla qubit measurement
        prob0_re = (bras_re * torch.matmul(meas_op, bras_re.conj().t()).t()).sum(dim=1).real
        prob0_im = (bras_im * torch.matmul(meas_op, bras_im.conj().t()).t()).sum(dim=1).real

    elif _is_counter_list(bras_re) and _is_counter_list(bras_im):
        # estimate overlap as the fraction of shots when 0 was observed on ancilla qubit
        n_qubits = len(list(bras_re[0].keys())[0])
        n_shots = sum(list(bras_re[0].values()))
        prob0_re = torch.tensor(
            [
                sum(map(lambda k, v: v if k[0] == "0" else 0, p.keys(), p.values())) / n_shots
                for p in bras_re
            ]
        )
        prob0_im = torch.tensor(
            [
                sum(map(lambda k, v: v if k[0] == "0" else 0, p.keys(), p.values())) / n_shots
                for p in bras_im
            ]
        )
    else:
        raise TypeError("Incorrect types passed for bras_re and kets_re arguments.")

    # construct final overlap tensor
    overlap = (2 * prob0_re - 1) ** 2 + (2 * prob0_im - 1) ** 2

    return overlap


class Overlap(QuantumModel):
    def __init__(
        self,
        bra_circuit: QuantumCircuit,
        ket_circuit: QuantumCircuit,
        backend: BackendName = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        measurement: Measurements | None = None,
        configuration: BackendConfiguration | dict | None = None,
        method: OverlapMethod = OverlapMethod.EXACT,
    ):
        self.backend_name = backend
        self.method = method

        overlap_fn, bra_circuit, ket_circuit = _select_overlap_method(
            method, backend, bra_circuit, ket_circuit
        )
        self.overlap_fn = overlap_fn

        super().__init__(
            bra_circuit,
            backend=backend,
            diff_mode=diff_mode,
            measurement=measurement,
            configuration=configuration,
        )
        self.bra_feat_param_names = set([inp.name for inp in self.inputs])

        if ket_circuit:
            self.ket_model = QuantumModel(
                ket_circuit,
                backend=backend,
                diff_mode=diff_mode,
                measurement=measurement,
                configuration=configuration,
            )
            self.ket_feat_param_names = set([inp.name for inp in self.ket_model.inputs])
        else:
            self.ket_model = None  # type: ignore [assignment]
            self.ket_feat_param_names = set([])

    def _process_param_values(
        self, bra_param_values: dict[str, Tensor], ket_param_values: dict[str, Tensor]
    ) -> dict:
        # we assume that either batch sizes are equal or 0 in case when no user params
        # are present in bra/ket
        bra_param_values = {
            k: v.reshape(-1) if v.shape == () else v for k, v in bra_param_values.items()
        }
        batch_size_bra = (
            len(list(bra_param_values.values())[0]) if len(bra_param_values) != 0 else 0
        )
        ket_param_values = {
            k: v.reshape(-1) if v.shape == () else v for k, v in ket_param_values.items()
        }
        batch_size_ket = (
            len(list(ket_param_values.values())[0]) if len(ket_param_values) != 0 else 0
        )
        new_bra_param_values = bra_param_values.copy()
        new_ket_param_values = ket_param_values.copy()

        # if len(self.bra_feat_param_names) + len(self.ket_feat_param_names) <= 2:

        if len(self.bra_feat_param_names.union(self.ket_feat_param_names)) == 2:
            # extend bra parameter tensors
            for param_name in new_bra_param_values.keys():
                new_bra_param_values[param_name] = new_bra_param_values[param_name].repeat(
                    batch_size_ket
                )

            # extend ket parameter tensors
            for param_name in new_ket_param_values.keys():
                idxs = torch.cat(
                    [
                        torch.ones(batch_size_bra, dtype=torch.int64) * i
                        for i in range(batch_size_ket)
                    ]
                )
                new_ket_param_values[param_name] = new_ket_param_values[param_name][idxs]

            if self.method in [OverlapMethod.EXACT, OverlapMethod.JENSEN_SHANNON]:
                param_values = {"bra": new_bra_param_values, "ket": new_ket_param_values}
            elif self.method in [
                OverlapMethod.COMPUTE_UNCOMPUTE,
                OverlapMethod.SWAP_TEST,
                OverlapMethod.HADAMARD_TEST,
            ]:
                # merge bra and ket param values to simulate all wavefunctions in one batch
                new_bra_param_values.update(new_ket_param_values)
                param_values = {"bra": new_bra_param_values}
                if self.method == OverlapMethod.HADAMARD_TEST:
                    param_values["ket"] = new_bra_param_values

        elif len(self.bra_feat_param_names.union(self.ket_feat_param_names)) < 2:
            if batch_size_bra == batch_size_ket or batch_size_bra == 0 or batch_size_ket == 0:
                param_values = {"bra": bra_param_values, "ket": ket_param_values}
            else:
                raise ValueError("Batch sizes of both bra and ket parameters must be equal.")

        else:
            raise ValueError("Multiple feature parameters for bra/ket are not currently supported.")

        return param_values

    def forward(  # type: ignore [override]
        self,
        bra_param_values: dict[str, Tensor] = {},
        ket_param_values: dict[str, Tensor] = {},
        bra_state: Tensor | None = None,
        ket_state: Tensor | None = None,
        n_shots: int = 0,
    ) -> Tensor:
        # reformat parameters
        param_values = self._process_param_values(bra_param_values, ket_param_values)

        # determine bra and ket calculation functions
        if n_shots == 0:
            bra_calc_fn = getattr(self, "run")
            ket_calc_fn = getattr(self.ket_model, "run", None)
        else:

            def bra_calc_fn(values: dict, state: Tensor) -> Any:
                return getattr(self, "sample")(values, n_shots, state)

            def ket_calc_fn(values: dict, state: Tensor) -> Any:
                return getattr(self.ket_model, "sample", lambda *_: _)(values, n_shots, state)

        # calculate overlap
        overlap = self.overlap_fn(
            param_values, bra_calc_fn, bra_state, ket_calc_fn, ket_state  # type: ignore [arg-type]
        )

        # reshape output if needed
        if len(self.bra_feat_param_names.union(self.ket_feat_param_names)) < 2:
            overlap = overlap[:, None]
        else:
            batch_size_bra = max(len(list(bra_param_values.values())[0]), 1)
            batch_size_ket = max(len(list(ket_param_values.values())[0]), 1)
            overlap = overlap.reshape((batch_size_ket, batch_size_bra)).t()

        return overlap



from __future__ import annotations

from .config import TrainConfig
from .data import DictDataLoader, InfiniteTensorDataset, to_dataloader
from .optimize_step import optimize_step as default_optimize_step
from .parameters import get_parameters, num_parameters, set_parameters
from .printing import print_metrics, write_tensorboard
from .saveload import load_checkpoint, load_model, write_checkpoint
from .tensors import numpy_to_tensor, promote_to, promote_to_tensor
from .train_grad import train as train_with_grad
from .train_no_grad import train as train_gradient_free

# Modules to be automatically added to the qadence namespace
__all__ = [
    "TrainConfig",
    "DictDataLoader",
    "train_with_grad",
    "train_gradient_free",
    "load_checkpoint",
    "write_checkpoint",
]



from __future__ import annotations

import torch

from qadence.blocks import AbstractBlock, add, chain, kron
from qadence.blocks.utils import block_is_qubit_hamiltonian
from qadence.constructors.hamiltonians import hamiltonian_factory
from qadence.logger import get_logger
from qadence.operations import HamEvo, I, N, X
from qadence.types import GenDAQC, Interaction, Strategy

from .gen_parser import _check_compatibility, _parse_generator
from .utils import _build_matrix_M, _ix_map

logger = get_logger(__name__)


def daqc_transform(
    n_qubits: int,
    gen_target: AbstractBlock,
    t_f: float,
    gen_build: AbstractBlock | None = None,
    zero_tol: float = 1e-08,
    strategy: Strategy = Strategy.SDAQC,
    ignore_global_phases: bool = False,
) -> AbstractBlock:
    """
    Implements the DAQC transform for representing an arbitrary 2-body Hamiltonian.

    The result is another fixed 2-body Hamiltonian.

    Reference for universality of 2-body Hamiltonians:

    -- https://arxiv.org/abs/quant-ph/0106064

    Based on the transformation for Ising (ZZ) interactions, as described in the paper

    -- https://arxiv.org/abs/1812.03637

    The transform translates a target weighted generator of the type:

        `gen_target = add(g_jk * kron(op(j), op(k)) for j < k)`

    To a circuit using analog evolutions with a fixed building block generator:

        `gen_build = add(f_jk * kron(op(j), op(k)) for j < k)`

    where `op = Z` or `op = N`.

    Args:
        n_qubits: total number of qubits to use.
        gen_target: target generator built with the structure above. The type
            of the generator will be automatically evaluated when parsing.
        t_f (float): total time for the gen_target evolution.
        gen_build: fixed generator to act as a building block. Defaults to
            constant NN: add(1.0 * kron(N(j), N(k)) for j < k). The type
            of the generator will be automatically evaluated when parsing.
        zero_tol: default "zero" for a missing interaction. Included for
            numerical reasons, see notes below.
        strategy: sDAQC or bDAQC, following definitions in the reference paper.
        ignore_global_phases: if `True` the transform does not correct the global
            phases coming from the mapping between ZZ and NN interactions.

    Notes:

        The paper follows an index convention of running from 1 to N. A few functions
        here also use that convention to be consistent with the paper. However, for qadence
        related things the indices are converted to [0, N-1].

        The case for `n_qubits = 4` is an edge case where the sign matrix is not invertible.
        There is a workaround for this described in the paper, but it is currently not implemented.

        The current implementation may result in evolution times that are both positive or
        negative. In practice, both can be represented by simply changing the signs of the
        interactions. However, for a real implementation where the interactions should remain
        fixed, the paper discusses a workaround that is not currently implemented.

        The transformation works by representing each interaction in the target hamiltonian by
        a set of evolutions using the build hamiltonian. As a consequence, some care must be
        taken when choosing the build hamiltonian. Some cases:

        - The target hamiltonian can have any interaction, as long as it is sufficiently
        represented in the build hamiltonian. E.g., if the interaction `g_01 * kron(Z(0), Z(1))`
        is in the target hamiltonian, the corresponding interaction `f_01 * kron(Z(0), Z(1))`
        needs to be in the build hamiltonian. This is checked when the generators are parsed.

        - The build hamiltonian can have any interaction, irrespectively of it being needed
        for the target hamiltonian. This is especially useful for designing local operations
        through the repeated evolution of a "global" hamiltonian.

        - The parameter `zero_tol` controls what it means for an interaction to be "missing".
        Any interaction strength smaller than `zero_tol` in the build hamiltonian will not be
        considered, and thus that interaction is missing.

        - The various ratios `g_jk / f_jk` will influence the time parameter for the various
        evolution slices, meaning that if there is a big discrepancy in the interaction strength
        for a given qubit pair (j, k), the output circuit may require the usage of hamiltonian
        evolutions with very large times.

        - A warning will be issued for evolution times larger than `1/sqrt(zero_tol)`. Evolution
        times smaller than `zero_tol` will not be represented.

    Examples:
        ```python exec="on" source="material-block" result="json"
        from qadence import Z, N, daqc_transform

        n_qubits = 3

        gen_build = 0.5 * (N(0)@N(1)) + 0.7 * (N(1)@N(2)) + 0.2 * (N(0)@N(2))

        gen_target = 0.1 * (Z(1)@Z(2))

        t_f = 2.0

        transformed_circuit = daqc_transform(
            n_qubits = n_qubits,
            gen_target = gen_target,
            t_f = t_f,
            gen_build = gen_build,
        )
        ```
    """

    ##################
    # Input controls #
    ##################

    if strategy != Strategy.SDAQC:
        raise NotImplementedError("Currently only the sDAQC transform is implemented.")

    if n_qubits == 4:
        raise NotImplementedError("DAQC transform 4-qubit edge case not implemented.")

    if gen_build is None:
        gen_build = hamiltonian_factory(n_qubits, interaction=Interaction.NN)

    try:
        if (not block_is_qubit_hamiltonian(gen_target)) or (
            not block_is_qubit_hamiltonian(gen_build)
        ):
            raise ValueError(
                "Generator block is not a qubit Hamiltonian. Only ZZ or NN interactions allowed."
            )
    except NotImplementedError:
        # Happens when block_is_qubit_hamiltonian is called on something that is not a block.
        raise TypeError(
            "Generator block is not a qubit Hamiltonian. Only ZZ or NN interactions allowed."
        )

    #####################
    # Generator parsing #
    #####################

    g_jk_target, mat_jk_target, target_type = _parse_generator(n_qubits, gen_target, 0.0)
    g_jk_build, mat_jk_build, build_type = _parse_generator(n_qubits, gen_build, zero_tol)

    # Get the global phase hamiltonian and single-qubit detuning hamiltonian
    if build_type == GenDAQC.NN:
        h_phase_build, h_sq_build = _nn_phase_and_detunings(n_qubits, mat_jk_build)

    if target_type == GenDAQC.NN:
        h_phase_target, h_sq_target = _nn_phase_and_detunings(n_qubits, mat_jk_target)

    # Time re-scalings
    if build_type == GenDAQC.ZZ and target_type == GenDAQC.NN:
        t_star = t_f / 4.0
    elif build_type == GenDAQC.NN and target_type == GenDAQC.ZZ:
        t_star = 4.0 * t_f
    else:
        t_star = t_f

    # Check if target Hamiltonian can be mapped with the build Hamiltonian
    assert _check_compatibility(g_jk_target, g_jk_build, zero_tol)

    ##################
    # DAQC Transform #
    ##################

    # Section III A of https://arxiv.org/abs/1812.03637:

    # Matrix M for the linear system, exemplified in Table I:
    matrix_M = _build_matrix_M(n_qubits)

    # Linear system mapping interaction ratios -> evolution times.
    t_slices = torch.linalg.solve(matrix_M, g_jk_target / g_jk_build) * t_star

    # ZZ-DAQC with ZZ or NN build Hamiltonian
    daqc_slices = []
    for m in range(2, n_qubits + 1):
        for n in range(1, m):
            alpha = _ix_map(n_qubits, n, m)
            t = t_slices[alpha - 1]
            if abs(t) > zero_tol:
                if abs(t) > (1 / (zero_tol**0.5)):
                    logger.warning(
                        """
Transformed circuit with very long evolution time.
Make sure your target interactions are sufficiently
represented in the build Hamiltonian."""
                    )
                x_gates = kron(X(n - 1), X(m - 1))
                analog_evo = HamEvo(gen_build, t)
                # TODO: Fix repeated X-gates
                if build_type == GenDAQC.NN:
                    # Local detuning at each DAQC layer for NN build Hamiltonian
                    sq_detuning_build = HamEvo(h_sq_build, t)
                    daqc_slices.append(chain(x_gates, sq_detuning_build, analog_evo, x_gates))
                elif build_type == GenDAQC.ZZ:
                    daqc_slices.append(chain(x_gates, analog_evo, x_gates))

    daqc_circuit = chain(*daqc_slices)

    ########################
    # Phases and Detunings #
    ########################

    if target_type == GenDAQC.NN:
        # Local detuning given a NN target Hamiltonian
        sq_detuning_target = HamEvo(h_sq_target, t_f).dagger()
        daqc_circuit = chain(sq_detuning_target, daqc_circuit)

    if not ignore_global_phases:
        if build_type == GenDAQC.NN:
            # Constant global phase given a NN build Hamiltonian
            global_phase_build = HamEvo(h_phase_build, t_slices.sum())
            daqc_circuit = chain(global_phase_build, daqc_circuit)

        if target_type == GenDAQC.NN:
            # Constant global phase and given a NN target Hamiltonian
            global_phase_target = HamEvo(h_phase_target, t_f).dagger()
            daqc_circuit = chain(global_phase_target, daqc_circuit)

    return daqc_circuit


def _nn_phase_and_detunings(
    n_qubits: int,
    mat_jk: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor]:
    # Constant global shift, leads to a global phase
    global_shift = mat_jk.sum() / 8

    # Strength of the local detunings
    g_sq = mat_jk.sum(0) / 2

    h_phase = global_shift * kron(I(i) for i in range(n_qubits))
    h_sq = add(-1.0 * g_sq[i] * N(i) for i in range(n_qubits))

    return h_phase, h_sq



from __future__ import annotations

from copy import deepcopy
from itertools import product
from math import dist
from typing import Any

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from deepdiff import DeepDiff
from networkx.classes.reportviews import EdgeView, NodeView

from qadence.analog import IdealDevice, RydbergDevice
from qadence.types import LatticeTopology

# Modules to be automatically added to the qadence namespace
__all__ = ["Register"]


DEFAULT_DEVICE = IdealDevice()


def _scale_node_positions(graph: nx.Graph, min_distance: float, spacing: float) -> None:
    scaled_nodes = {}
    scale_factor = spacing / min_distance
    for k, node in graph.nodes.items():
        (x, y) = node["pos"]
        scaled_nodes[k] = {"pos": (x * scale_factor, y * scale_factor)}
    nx.set_node_attributes(graph, scaled_nodes)


class Register:
    def __init__(
        self,
        support: nx.Graph | int,
        spacing: float | None = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ):
        """
        A 2D register of qubits which includes their coordinates.

        It is needed for e.g. analog computing.
        The coordinates are ignored in backends that don't need them. The easiest
        way to construct a register is via its classmethods like `Register.triangular_lattice`.

        Arguments:
            support: A graph or number of qubits. Nodes can include a `"pos"` attribute
                such that e.g.: `graph.nodes = {0: {"pos": (2,3)}, 1: {"pos": (0,0)}, ...}` which
                will be used in backends that need qubit coordinates.
                See the classmethods for simple construction of some predefined lattices if you
                don't want to build a graph manually.
                If you pass an integer the resulting register is the same as
                `Register.all_to_all(n_qubits)`.
            spacing: Value set as the distance between the two closest qubits.

        Examples:
        ```python exec="on" source="material-block" result="json"
        from qadence import Register

        reg = Register.honeycomb_lattice(2,3)
        reg.draw()
        ```
        """
        if device_specs is not None and not isinstance(device_specs, RydbergDevice):
            raise ValueError("Device specs are not valid. Please pass a `RydbergDevice` instance.")

        self.device_specs = device_specs

        self.graph = support if isinstance(support, nx.Graph) else alltoall_graph(support)

        if spacing is not None and self.min_distance != 0.0:
            _scale_node_positions(self.graph, self.min_distance, spacing)

    @property
    def n_qubits(self) -> int:
        return len(self.graph)

    @classmethod
    def from_coordinates(
        cls,
        coords: list[tuple],
        lattice: LatticeTopology | str = LatticeTopology.ARBITRARY,
        spacing: float | None = None,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        graph = nx.Graph()
        for i, pos in enumerate(coords):
            graph.add_node(i, pos=pos)
        return cls(graph, spacing, device_specs)

    @classmethod
    def line(
        cls,
        n_qubits: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        return cls(line_graph(n_qubits), spacing, device_specs)

    @classmethod
    def circle(
        cls,
        n_qubits: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        graph = nx.grid_2d_graph(n_qubits, 1, periodic=True)
        graph = nx.relabel_nodes(graph, {(i, 0): i for i in range(n_qubits)})
        coords = nx.circular_layout(graph)
        values = {i: {"pos": pos} for i, pos in coords.items()}
        nx.set_node_attributes(graph, values)
        return cls(graph, spacing, device_specs)

    @classmethod
    def square(
        cls,
        qubits_side: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        n_points = 4 * (qubits_side - 1)

        def gen_points() -> np.ndarray:
            rotate_left = np.array([[0.0, -1.0], [1.0, 0.0]])
            increment = np.array([0.0, 1.0])

            points = [np.array([0.0, 0.0])]
            counter = 1
            while len(points) < n_points:
                points.append(points[-1] + increment)

                counter = (counter + 1) % qubits_side
                if counter == 0:
                    increment = rotate_left.dot(increment)
                    counter = 1
            points = np.array(points)  # type: ignore[assignment]
            points -= np.mean(points, axis=0)

            return points  # type: ignore[return-value]

        graph = nx.grid_2d_graph(n_points, 1, periodic=True)
        graph = nx.relabel_nodes(graph, {(i, 0): i for i in range(n_points)})
        values = {i: {"pos": point} for i, point in zip(graph.nodes, gen_points())}
        nx.set_node_attributes(graph, values)
        return cls(graph, spacing, device_specs)

    @classmethod
    def all_to_all(
        cls,
        n_qubits: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        return cls(alltoall_graph(n_qubits), spacing, device_specs)

    @classmethod
    def rectangular_lattice(
        cls,
        qubits_row: int,
        qubits_col: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        graph = nx.grid_2d_graph(qubits_col, qubits_row)
        values = {i: {"pos": node} for (i, node) in enumerate(graph.nodes)}
        graph = nx.relabel_nodes(graph, {(i, j): k for k, (i, j) in enumerate(graph.nodes)})
        nx.set_node_attributes(graph, values)
        return cls(graph, spacing, device_specs)

    @classmethod
    def triangular_lattice(
        cls,
        n_cells_row: int,
        n_cells_col: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        return cls(triangular_lattice_graph(n_cells_row, n_cells_col), spacing, device_specs)

    @classmethod
    def honeycomb_lattice(
        cls,
        n_cells_row: int,
        n_cells_col: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        graph = nx.hexagonal_lattice_graph(n_cells_row, n_cells_col)
        graph = nx.relabel_nodes(graph, {(i, j): k for k, (i, j) in enumerate(graph.nodes)})
        return cls(graph, spacing, device_specs)

    @classmethod
    def lattice(cls, topology: LatticeTopology | str, *args: Any, **kwargs: Any) -> Register:
        return getattr(cls, topology)(*args, **kwargs)  # type: ignore[no-any-return]

    def draw(self, show: bool = True) -> None:
        coords = {i: n["pos"] for i, n in self.graph.nodes.items()}
        nx.draw(self.graph, with_labels=True, pos=coords)
        if show:
            plt.gcf().show()

    def __getitem__(self, item: int) -> Any:
        return self.graph.nodes[item]

    @property
    def nodes(self) -> NodeView:
        return self.graph.nodes

    @property
    def edges(self) -> EdgeView:
        return self.graph.edges

    @property
    def support(self) -> set:
        return set(self.nodes)

    @property
    def coords(self) -> dict:
        return {i: tuple(node.get("pos", ())) for i, node in self.nodes.items()}

    @property
    def all_node_pairs(self) -> EdgeView:
        return list(filter(lambda x: x[0] < x[1], product(self.support, self.support)))

    @property
    def distances(self) -> dict:
        coords = self.coords
        return {edge: dist(coords[edge[0]], coords[edge[1]]) for edge in self.all_node_pairs}

    @property
    def edge_distances(self) -> dict:
        coords = self.coords
        return {edge: dist(coords[edge[0]], coords[edge[1]]) for edge in self.edges}

    @property
    def min_distance(self) -> float:
        distances = self.distances
        value: float = min(self.distances.values()) if len(distances) > 0 else 0.0
        return value

    def rescale_coords(self, scaling: float) -> Register:
        g = deepcopy(self.graph)
        _scale_node_positions(g, min_distance=1.0, spacing=scaling)
        return Register(g, spacing=None, device_specs=self.device_specs)

    def _to_dict(self) -> dict:
        return {
            "graph": nx.node_link_data(self.graph),
            "device_specs": self.device_specs._to_dict(),
        }

    @classmethod
    def _from_dict(cls, d: dict) -> Register:
        device_dict = d.get("device_specs", None)
        if device_dict is None:
            device_dict = DEFAULT_DEVICE._to_dict()

        return cls(
            support=nx.node_link_graph(d["graph"]),
            device_specs=RydbergDevice._from_dict(device_dict),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Register):
            return False
        return (
            DeepDiff(self.coords, other.coords, ignore_order=False) == {}
            and nx.is_isomorphic(self.graph, other.graph)
            and self.n_qubits == other.n_qubits
        )


def line_graph(n_qubits: int) -> nx.Graph:
    """Create graph representing linear lattice.

    Args:
        n_qubits (int): number of nodes in the graph

    Returns:
        graph instance
    """
    graph = nx.Graph()
    for i in range(n_qubits):
        graph.add_node(i, pos=(i, 0.0))
    for i, j in zip(range(n_qubits - 1), range(1, n_qubits)):
        graph.add_edge(i, j)
    return graph


def triangular_lattice_graph(n_cells_row: int, n_cells_col: int) -> nx.Graph:
    graph = nx.triangular_lattice_graph(n_cells_row, n_cells_col)
    graph = nx.relabel_nodes(graph, {(i, j): k for k, (i, j) in enumerate(graph.nodes)})
    return graph


def alltoall_graph(n_qubits: int) -> nx.Graph:
    if n_qubits == 2:
        return line_graph(2)
    if n_qubits == 3:
        return triangular_lattice_graph(1, 1)

    graph = nx.complete_graph(n_qubits)
    # set seed to make sure the produced graphs are reproducible
    coords = nx.spring_layout(graph, seed=0)
    for i, pos in coords.items():
        graph.nodes[i]["pos"] = tuple(pos)
    return graph



from __future__ import annotations

import numpy as np
import sympy
import torch
from torch import Tensor, cdouble

from qadence.blocks import (
    AbstractBlock,
    ParametricBlock,
)
from qadence.blocks.utils import (
    add,  # noqa
    chain,
)
from qadence.logger import get_logger
from qadence.parameters import (
    Parameter,
    ParamMap,
    evaluate,
)
from qadence.types import OpName, TNumber, TParameter

from .primitive import I, X, Y, Z

logger = get_logger(__name__)


class PHASE(ParametricBlock):
    """The Parametric Phase / S gate."""

    name = OpName.PHASE

    def __init__(self, target: int, parameter: Parameter | TNumber | sympy.Expr | str):
        self.parameters = ParamMap(parameter=parameter)
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 2], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        lmbda = torch.exp(1j * evaluate(self.parameters.parameter, as_torch=True))
        return torch.cat((torch.ones(1), lmbda))


class RX(ParametricBlock):
    """The Rx gate."""

    name = OpName.RX

    def __init__(self, target: int, parameter: Parameter | TParameter | ParamMap):
        # TODO: should we give them more meaningful names? like 'angle'?
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        self.generator = X(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((lmbd, lmbd.conj()))


class RY(ParametricBlock):
    """The Ry gate."""

    name = OpName.RY

    def __init__(self, target: int, parameter: Parameter | TParameter | ParamMap):
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        self.generator = Y(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((lmbd, lmbd.conj()))


class RZ(ParametricBlock):
    """The Rz gate."""

    name = OpName.RZ

    def __init__(self, target: int, parameter: Parameter | TParameter | ParamMap):
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        self.generator = Z(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((lmbd, lmbd.conj()))


class U(ParametricBlock):
    """Arbitrary one-qubit rotation in the Bloch sphere.

    This operation accepts 3 parameters (phi, theta, omega)
    """

    name = OpName.U

    def __init__(
        self,
        target: int,
        phi: Parameter | TParameter,
        theta: Parameter | TParameter,
        omega: Parameter | TParameter,
    ):
        self.parameters = ParamMap(phi=phi, theta=theta, omega=omega)
        self.generator = chain(Z(target), Y(target), Z(target))
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 3

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=torch.cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        phi = evaluate(self.parameters.phi)
        theta = evaluate(self.parameters.theta)
        omega = evaluate(self.parameters.omega)
        lmbd = np.exp(-1j * (phi + omega) / 2) * np.cos(theta / 2)
        return torch.cat((lmbd, lmbd.conj()))

    @property
    def n_qubits(self) -> int:
        return 1

    def digital_decomposition(self) -> AbstractBlock:
        return chain(
            RZ(self.qubit_support[0], self.parameters.phi),
            RY(self.qubit_support[0], self.parameters.theta),
            RZ(self.qubit_support[0], self.parameters.omega),
        )



from __future__ import annotations

from typing import Any, ItemsView, KeysView, ValuesView, get_args
from uuid import uuid4

import jsonschema
import numpy as np
import sympy
from sympy import *
from sympy import Array, Basic, Expr, Symbol, sympify
from sympy.physics.quantum.dagger import Dagger
from sympytorch import SymPyModule as torchSympyModule
from torch import Tensor, heaviside, no_grad, rand, tensor

from qadence.logger import get_logger
from qadence.types import DifferentiableExpression, Engine, TNumber

# Modules to be automatically added to the qadence namespace
__all__ = ["FeatureParameter", "Parameter", "VariationalParameter"]

logger = get_logger(__file__)

dagger_expression = Dagger

ParameterJSONSchema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "trainable": {"type": "string"},
        "value": {"type": "string"},
    },
    "oneOf": [
        {
            "allOf": [
                {"required": ["name"]},
                {"not": {"required": ["trainable"]}},
                {"not": {"required": ["value"]}},
            ]
        },
        {"allOf": [{"required": ["name", "trainable"]}, {"not": {"required": ["value"]}}]},
        {"required": ["name", "trainable", "value"]},
    ],
}


class Parameter(Symbol):
    """
    A wrapper on top of `sympy.Symbol`.

    Includes two additional keywords: `trainable` and `value`.
    This class is to define both feature parameter and variational parameters.
    """

    trainable: bool
    """Trainable parameters are *variational* parameters.

    Non-trainable parameters are *feature*
    parameters.
    """
    value: TNumber
    """(Initial) value of the parameter."""

    def __new__(
        cls, name: str | TNumber | Tensor | Basic | Parameter, **assumptions: Any
    ) -> Parameter | Basic | Expr | Array:
        """
        Arguments:

            name: When given a string only, the class
                constructs a trainable Parameter with a a randomly initialized value.
            **assumptions: are passed on to the parent class `sympy.Symbol`. Two new assumption
                kwargs are supported by this constructor: `trainable: bool`, and `value: TNumber`.

        Example:
        ```python exec="on" source="material-block" result="json"
        from qadence.parameters import Parameter, VariationalParameter

        theta = Parameter("theta")
        print(f"{theta}: trainable={theta.trainable} value={theta.value}")
        assert not theta.is_number

        # you can specify both trainable/value in the constructor
        theta = Parameter("theta", trainable=True, value=2.0)
        print(f"{theta}: trainable={theta.trainable} value={theta.value}")

        # VariationalParameter/FeatureParameter are constructing
        # trainable/untrainable Parameters
        theta = VariationalParameter("theta", value=2.0)
        assert theta == Parameter("theta", trainable=True, value=2.0)

        # When provided with a numeric type, Parameter constructs a sympy numeric type":
        constant_zero = Parameter(0)
        assert constant_zero.is_number

        # When passed a Parameter or a sympy expression, it just returns it.
        expr = Parameter("x") * Parameter("y")
        print(f"{expr=} : {expr.free_symbols}")
        ```
        """
        p: Parameter
        if isinstance(name, get_args(TNumber)):
            return sympify(name)
        elif isinstance(name, Tensor):
            if name.numel() == 1:
                return sympify(name)
            else:
                return Array(name.detach().numpy())
        elif isinstance(name, Parameter):
            p = super().__new__(cls, name.name, **assumptions)
            p.name = name.name
            p.trainable = name.trainable
            p.value = name.value
            return p
        elif isinstance(name, (Basic, Expr)):
            if name.is_number:
                return sympify(evaluate(name))
            return name
        elif isinstance(name, str):
            p = super().__new__(cls, name, **assumptions)
            p.trainable = assumptions.get("trainable", True)
            p.value = assumptions.get("value", None)
            if p.value is None:
                p.value = rand(1).item()
            return p
        else:
            raise TypeError(f"Parameter does not support type {type(name)}")

    def __eq__(self, other: object) -> bool:
        from qadence.utils import isclose

        if isinstance(other, str):
            return self.name == other  # type: ignore[no-any-return]

        elif isinstance(other, Parameter):
            return (
                self.name == other.name
                and self.trainable == other.trainable
                and isclose(self.value, other.value)
            )
        elif isinstance(other, Expr):
            return self in other.free_symbols
        elif isinstance(other, Symbol):
            return self.name == other.name  # type: ignore[no-any-return]

        return False

    def __hash__(self) -> Any:
        return super().__hash__()

    def _to_dict(self) -> dict:
        d = {"name": self.name, "trainable": str(self.trainable), "value": str(self.value)}
        try:
            jsonschema.validate(d, ParameterJSONSchema)
            return d
        except jsonschema.exceptions.ValidationError as e:
            logger.exception(f"Parameter dict {d} doesnt comply to {ParameterJSONSchema} with {e}.")
            return {}

    @classmethod
    def _from_dict(cls, d: dict) -> Parameter | None:
        try:
            jsonschema.validate(d, ParameterJSONSchema)
            trainable = True if d["trainable"] == "True" else False
            return cls(name=d["name"], trainable=trainable, value=float(d["value"]))
        except jsonschema.exceptions.ValidationError as e:
            logger.exception(f"Parameter dict {d} doesnt comply to {ParameterJSONSchema} with {e}.")
            return None


def FeatureParameter(name: str, **kwargs: Any) -> Parameter:
    """Shorthand for `Parameter(..., trainable=False)`."""
    return Parameter(name, trainable=False, **kwargs)


def VariationalParameter(name: str, **kwargs: Any) -> Parameter:
    """Shorthand for `Parameter(..., trainable=True)`."""
    return Parameter(name, trainable=True, **kwargs)


def extract_original_param_entry(
    param: Expr,
) -> TNumber | Tensor | Expr:
    """
    Given an Expression, what was the original "param" given by the user? It is either.

    going to be a numeric value, or a sympy Expression (in case a string was given,
    it was converted via Parameter("string").
    """
    return param if not param.is_number else evaluate(param)


def heaviside_func(x: Tensor, _: Any) -> Tensor:
    with no_grad():
        res = heaviside(x, tensor(0.5))
    return res


def torchify(expr: Expr) -> torchSympyModule:
    extra_funcs = {sympy.core.numbers.ImaginaryUnit: 1.0j, sympy.Heaviside: heaviside_func}
    return torchSympyModule(expressions=[sympy.N(expr)], extra_funcs=extra_funcs)


def make_differentiable(expr: Expr, engine: Engine = Engine.TORCH) -> DifferentiableExpression:
    diff_expr: DifferentiableExpression
    if engine == Engine.JAX:
        from qadence.backends.jax_utils import jaxify

        diff_expr = jaxify(expr)
    else:
        diff_expr = torchify(expr)
    return diff_expr


def sympy_to_numeric(expr: Basic) -> TNumber:
    if expr.as_real_imag()[1] != 0:
        return complex(expr)
    else:
        return float(expr)


def evaluate(expr: Expr, values: dict = {}, as_torch: bool = False) -> TNumber | Tensor:
    """
    Arguments:

        expr: An expression consisting of Parameters.
        values: values dict which contains values for the Parameters,
            if empty, Parameter.value will be used.
        as_torch: Whether to retrieve a torch-differentiable expression result.

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence.parameters import Parameter, evaluate

    expr = Parameter("x") * Parameter("y")

    # Unless specified, Parameter initialized random values
    # Lets evaluate this expression and see what the result is
    res = evaluate(expr)
    print(res)

    # We can also evaluate the expr using a custom dict
    d = {"x": 1, "y":2}
    res = evaluate(expr, d)
    print(res)

    # Lastly, if we want a differentiable result, lets put the as_torch flag
    res = evaluate(expr, d, as_torch=True)
    print(res)
    ```
    """
    res: Basic
    res_value: TNumber | Tensor
    query: dict[Parameter, TNumber | Tensor] = {}
    if isinstance(expr, Array):
        return Tensor(expr.tolist())
    else:
        if not expr.is_number:
            for s in expr.free_symbols:
                if s.name in values.keys():
                    query[s] = values[s.name]
                elif hasattr(s, "value"):
                    query[s] = s.value
                else:
                    raise ValueError(f"No value provided for symbol {s.name}")
        if as_torch:
            res_value = make_differentiable(expr)(**{s.name: tensor(v) for s, v in query.items()})
        else:
            res = expr.subs(query)
            res_value = sympy_to_numeric(res)
        return res_value


def stringify(expr: Basic) -> str:
    name: str = ""
    if isinstance(expr, Array):
        return str(np.array(expr.tolist())).replace(".", "_")
    else:
        if expr.is_number:
            expr_hash = hash(sympy_to_numeric(expr))
            name = "fix_" + str(expr_hash)
        else:
            name = str(expr).replace(".", "_")
        return name


class ParamMap:
    """Connects UUIDs of parameters to their expressions and names.

    This class is not user-facing
    and only needed for more complex block definitions. It provides convenient access to
    expressions/UUIDs/names needed in different backends.

    Arguments:
        kwargs: Parameters.

    Example:
    ```python exec="on" source="material-block" result="json"
    import sympy
    from qadence.parameters import ParamMap

    (x,y) = sympy.symbols("x y")
    ps = ParamMap(omega=2.0, duration=x+y)

    print(f"{ps.names() = }")
    print(f"{ps.expressions() = }")
    print(f"{ps.uuids() = }")
    ```
    """

    def __init__(self, **kwargs: str | TNumber | Tensor | Basic | Parameter):
        self._name_dict: dict[str, tuple[str, Basic]] = {}
        self._uuid_dict: dict[str, str] = {}
        for name, v in kwargs.items():
            param = v if isinstance(v, sympy.Basic) else Parameter(v)
            uuid = str(uuid4())
            self._name_dict[name] = (uuid, param)
            self._uuid_dict[uuid] = param

    def __getattr__(self, name: str) -> Basic:
        _name_dict = self.__getattribute__("_name_dict")
        if name in _name_dict:
            (_, param) = _name_dict[name]
            return param
        else:
            return self.__getattribute__(name)

    def uuid(self, name: str) -> str:
        (_uuid, _) = self._name_dict[name]
        return _uuid

    def param_str(self, name: str) -> str:
        return stringify(self.param(name))

    def uuid_param(self, name: str) -> tuple[str, Basic]:
        return self._name_dict[name]

    def names(self) -> KeysView:
        return self._name_dict.keys()

    def uuids(self) -> KeysView:
        return self._uuid_dict.keys()

    def expressions(self) -> ValuesView:
        return self._uuid_dict.values()

    def items(self) -> ItemsView:
        return self._uuid_dict.items()

    def __repr__(self) -> str:
        s = repr(self._name_dict)
        s = s.replace("{", "(")
        s = s.replace("}", ")")
        return "ParamMap" + s

    def _to_dict(self) -> dict:
        from qadence.serialization import serialize

        d = {name: (uuid, serialize(expr)) for (name, (uuid, expr)) in self._name_dict.items()}
        return {"_name_dict": d}

    @classmethod
    def _from_dict(cls, d: dict) -> ParamMap:
        from qadence.serialization import deserialize

        res = ParamMap()
        for name, (uuid, v) in d["_name_dict"].items():
            param: Parameter = deserialize(v)  # type: ignore[assignment]
            res._name_dict[name] = (uuid, param)
            res._uuid_dict[uuid] = param
        return res



from __future__ import annotations

import math
from collections import Counter
from typing import Any

import numpy as np
import sympy
from torch import Tensor, stack, vmap
from torch import complex as make_complex
from torch.linalg import eigvals

from qadence.logger import get_logger
from qadence.types import Endianness, ResultType, TNumber

# Modules to be automatically added to the qadence namespace
__all__ = []  # type: ignore


logger = get_logger(__name__)


def basis_to_int(basis: str, endianness: Endianness = Endianness.BIG) -> int:
    """
    Converts a computational basis state to an int.

    - `endianness = "Big"` reads the most significant bit in qubit 0 (leftmost).
    - `endianness = "Little"` reads the least significant bit in qubit 0 (leftmost).

    Arguments:
        basis (str): A computational basis state.
        endianness (Endianness): The Endianness when reading the basis state.

    Returns:
        The corresponding integer.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.utils import basis_to_int, Endianness

    k = basis_to_int(basis="10", endianness=Endianness.BIG)
    print(k)
    ```
    """
    if endianness == Endianness.BIG:
        return int(basis, 2)
    else:
        return int(basis[::-1], 2)


def int_to_basis(
    k: int, n_qubits: int | None = None, endianness: Endianness = Endianness.BIG
) -> str:
    """
    Converts an integer to its corresponding basis state.

    - `endianness = "Big"` stores the most significant bit in qubit 0 (leftmost).
    - `endianness = "Little"` stores the least significant bit in qubit 0 (leftmost).

    Arguments:
        k (int): The int to convert.
        n_qubits (int): The total number of qubits in the basis state.
        endianness (Endianness): The Endianness of the resulting basis state.

    Returns:
        A computational basis state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.utils import int_to_basis, Endianness

    bs = int_to_basis(k=1, n_qubits=2, endianness=Endianness.BIG)
    print(bs)
    ```
    """
    if n_qubits is None:
        n_qubits = int(math.log(k + 0.6) / math.log(2)) + 1
    assert k <= 2**n_qubits - 1, "k can not be larger than 2**n_qubits-1."
    basis = format(k, "0{}b".format(n_qubits))
    if endianness == Endianness.BIG:
        return basis
    else:
        return basis[::-1]


def nqubits_to_basis(
    n_qubits: int,
    result_type: ResultType = ResultType.STRING,
    endianness: Endianness = Endianness.BIG,
) -> list[str] | Tensor | np.array:
    """
    Creates all basis states for a given number of qubits, endianness and format.

    Arguments:
        n_qubits: The total number of qubits.
        result_type: The data type of the resulting states.
        endianness: The Endianness of the resulting states.

    Returns:
        The full computational basis for n_qubits.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.utils import nqubits_to_basis, Endianness, ResultType
    basis_type = ResultType.Torch
    bs = nqubits_to_basis(n_qubits=2, result_type= basis_type, endianness=Endianness.BIG)
    print(bs)
    ```
    """
    basis_strings = [int_to_basis(k, n_qubits, endianness) for k in range(0, 2**n_qubits)]
    if result_type == ResultType.STRING:
        return basis_strings
    else:
        basis_list = [list(map(int, tuple(basis))) for basis in basis_strings]
        if result_type == ResultType.TORCH:
            return stack([Tensor(basis) for basis in basis_list])
        elif result_type == ResultType.NUMPY:
            return np.stack([np.array(basis) for basis in basis_list])


def samples_to_integers(samples: Counter, endianness: Endianness = Endianness.BIG) -> Counter:
    """
    Converts a Counter of basis state samples to integer values.

    Args:
        samples (Counter({bits: counts})): basis state sample counter.
        endianness (Endianness): endianness to use for conversion.

    Returns:
        Counter({ints: counts}): samples converted
    """

    return Counter({basis_to_int(k, endianness): v for k, v in samples.items()})


def format_number(x: float | complex, num_digits: int = 3) -> str:
    if isinstance(x, int):
        return f"{x}"
    elif isinstance(x, float):
        return f"{x:.{num_digits}f}"
    elif isinstance(x, complex):
        re = "" if np.isclose(x.real, 0) else f"{x.real:.{num_digits}f}"
        im = "" if np.isclose(x.imag, 0) else f"{x.imag:.{num_digits}f}"
        if len(re) > 0 and len(im) > 0:
            return f"{re}+{im}j"
        elif len(re) > 0 and len(im) == 0:
            return re
        elif len(re) == 0 and len(im) > 0:
            return f"{im}j"
        else:
            return "0"
    else:
        raise ValueError(f"Unknown number type: {type(x)}")


def format_parameter(p: sympy.Basic) -> str:
    def round_expr(expr: sympy.Basic, num_digits: int) -> sympy.Basic:
        return expr.xreplace({n: round(n, num_digits) for n in expr.atoms(sympy.Number)})

    return str(round_expr(p, 3))


def print_sympy_expr(expr: sympy.Expr, num_digits: int = 3) -> str:
    """
    Converts numerical values in a sympy expression.

    The result is a numerical expression with fewer digits for better readability.
    """
    from qadence.parameters import sympy_to_numeric

    round_dict = {sympy_to_numeric(n): round(n, num_digits) for n in expr.atoms(sympy.Number)}
    return str(expr.xreplace(round_dict))


def isclose(
    x: TNumber | Any, y: TNumber | Any, rel_tol: float = 1e-5, abs_tol: float = 1e-07
) -> bool:
    if isinstance(x, complex) or isinstance(y, complex):
        return abs(x - y) <= max(rel_tol * max(abs(x), abs(y)), abs_tol)  # type: ignore

    return math.isclose(x, y, rel_tol=rel_tol, abs_tol=abs_tol)  # type: ignore


def eigenvalues(
    x: Tensor, max_num_evals: int | None = None, max_num_gaps: int | None = None
) -> Tensor:
    # FIXME: Currently doing full decomposition everytime because it is generally faster
    # than converting to numpy and using the scipy routines to select first k-eigenvalues.
    # Furthermore, we should exploit the matrices being Hermitian.

    # get all eigenvalues of generator
    eigenspectrum = eigvals(x).squeeze(0)

    if max_num_evals and not max_num_gaps:
        # get specified number of eigenvalues of generator
        eigenvals = eigenspectrum[:max_num_evals]
    elif max_num_gaps and not max_num_evals:
        # get eigenvalues of generator corresponding to specified number of spectral gaps
        k = int(np.ceil(0.5 * (1 + np.sqrt(1 + 8 * max_num_gaps))))
        eigenvals = eigenspectrum[:k]
    else:
        eigenvals = eigenspectrum

    return eigenvals


def _round_complex(t: Tensor, decimals: int = 4) -> Tensor:
    def _round(_t: Tensor) -> Tensor:
        r = _t.real.round(decimals=decimals)
        i = _t.imag.round(decimals=decimals)
        return make_complex(r, i)

    fn = vmap(_round)
    return fn(t)


def is_qadence_shape(state: Tensor, n_qubits: int) -> bool:
    if len(state.size()) < 2:
        raise ValueError(
            f"Provided state is required to have atleast two dimensions. Got shape {state.shape}"
        )
    return state.shape[1] == 2**n_qubits  # type: ignore[no-any-return]


def infer_batchsize(param_values: dict[str, Tensor] = None) -> int:
    """Infer the batch_size through the length of the parameter tensors."""
    try:
        return max([len(tensor) for tensor in param_values.values()]) if param_values else 1
    except Exception:
        return 1


def validate_values_and_state(
    state: Tensor | None, n_qubits: int, param_values: dict[str, Tensor] = None
) -> None:
    if state is not None:
        batch_size_state = (
            state.shape[0] if is_qadence_shape(state, n_qubits=n_qubits) else state.size(-1)
        )
        batch_size_values = infer_batchsize(param_values)
        if batch_size_state != batch_size_values and (
            batch_size_values > 1 and batch_size_state > 1
        ):
            raise ValueError(
                "Batching of parameter values and states is only valid for the cases:\
                            (1) batch_size_values == batch_size_state\
                            (2) batch_size_values == 1 and batch_size_state > 1\
                            (3) batch_size_values > 1 and batch_size_state == 1."
            )



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import Any

import numpy as np
import qutip
import torch
from pulser import Register as PulserRegister
from pulser import Sequence
from pulser_simulation import SimConfig
from pulser_simulation.simresults import SimulationResults
from pulser_simulation.simulation import QutipEmulator
from torch import Tensor

from qadence.backend import Backend as BackendInterface
from qadence.backend import ConvertedCircuit, ConvertedObservable
from qadence.backends.utils import to_list_of_dicts
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.logger import get_logger
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.mitigations.protocols import apply_mitigation
from qadence.noise import Noise
from qadence.noise.protocols import apply_noise
from qadence.overlap import overlap_exact
from qadence.register import Register
from qadence.transpile import transpile
from qadence.types import BackendName, DeviceType, Endianness, Engine

from .channels import GLOBAL_CHANNEL, LOCAL_CHANNEL
from .cloud import get_client
from .config import Configuration
from .convert_ops import convert_observable
from .devices import IdealDevice, RealisticDevice
from .pulses import add_addressing_pattern, add_pulses

logger = get_logger(__file__)


def _convert_init_state(state: Tensor) -> np.ndarray:
    """Flip and squeeze initial state consistent with Pulser convention."""
    if state.shape[0] > 1:
        raise ValueError("Pulser backend only supports initial states with batch size 1.")
    return np.flip(state.cpu().squeeze().numpy())


def create_register(register: Register) -> PulserRegister:
    """Convert Qadence Register to Pulser Register."""
    coords = np.array(list(register.coords.values()))
    return PulserRegister.from_coordinates(coords)


def make_sequence(circ: QuantumCircuit, config: Configuration) -> Sequence:
    qadence_register = circ.register
    device_specs = qadence_register.device_specs

    if device_specs.type == DeviceType.IDEALIZED:
        device = IdealDevice(
            device_specs.rydberg_level, device_specs.max_detuning, device_specs.max_amp
        )
    elif device_specs.type == DeviceType.REALISTIC:
        device = RealisticDevice(
            device_specs.rydberg_level, device_specs.max_detuning, device_specs.max_amp
        )
    else:
        raise ValueError(
            f"Specified device of type {device_specs.type} is not supported by the pulser backend."
        )

    ########
    # FIXME: Remove the block below in V1.1.0
    if config.spacing is not None:
        logger.warning(
            "Passing register spacing in the backend configuration is deprecated. "
            "Please pass it in the register directly, as detailed in the register tutorial."
        )
        # Rescales the register coordinates, as was done with the previous "spacing" argument.
        qadence_register = qadence_register.rescale_coords(scaling=config.spacing)
    else:
        if qadence_register.min_distance < 4.0:
            # Throws warning for minimum distance below 4 because the typical values used
            # for the standard pulser device parameters is ~7-8, so this likely means the user
            # forgot to set the spacing at register creation.
            logger.warning(
                "Register with distance between atoms smaller than 4µm detected. "
                "Pulser backend no longer has a default spacing of 8µm applied to the register. "
                "Make sure you set the desired spacing as detailed in the register tutorial."
            )
    ########

    pulser_register = create_register(qadence_register)

    sequence = Sequence(pulser_register, device)

    sequence.declare_channel(GLOBAL_CHANNEL, "rydberg_global")
    sequence.declare_channel(LOCAL_CHANNEL, "rydberg_local", initial_target=0)

    add_pulses(sequence, circ.block, config, qadence_register)

    return sequence


# TODO: make it parallelized
def simulate_sequence(
    sequence: Sequence, config: Configuration, state: Tensor, n_shots: int | None = None
) -> SimulationResults | Counter:
    if config.cloud_configuration is not None:
        client = get_client(config.cloud_configuration)

        serialized_sequence = sequence.to_abstract_repr()
        params: list[dict] = [{"runs": n_shots, "variables": {}}]

        batch = client.create_batch(
            serialized_sequence,
            jobs=params,
            emulator=str(config.cloud_configuration.platform),
            wait=True,
        )

        job = list(batch.jobs.values())[0]
        if job.errors is not None:
            logger.error(
                f"The cloud job with ID {job.id} has "
                f"failed for the following reason: {job.errors}"
            )

        return Counter(job.result)

    else:
        simulation = QutipEmulator.from_sequence(
            sequence,
            sampling_rate=config.sampling_rate,
            config=config.sim_config,
            with_modulation=config.with_modulation,
        )
        if state is not None:
            simulation.set_initial_state(qutip.Qobj(state))

        sim_result = simulation.run(nsteps=config.n_steps_solv, method=config.method_solv)
        if n_shots is not None:
            return sim_result.sample_final_state(n_shots)
        else:
            return sim_result


@dataclass(frozen=True, eq=True)
class Backend(BackendInterface):
    """The Pulser backend."""

    name: BackendName = BackendName.PULSER
    supports_ad: bool = False
    supports_adjoint: bool = False
    support_bp: bool = False
    is_remote: bool = False
    with_measurements: bool = True
    with_noise: bool = False
    native_endianness: Endianness = Endianness.BIG
    config: Configuration = field(default_factory=Configuration)
    engine: Engine = Engine.TORCH

    def circuit(self, circ: QuantumCircuit) -> Sequence:
        passes = self.config.transpilation_passes
        original_circ = circ
        if passes is not None and len(passes) > 0:
            circ = transpile(*passes)(circ)

        native = make_sequence(circ, self.config)

        return ConvertedCircuit(native=native, abstract=circ, original=original_circ)

    def observable(self, observable: AbstractBlock, n_qubits: int = None) -> Tensor:
        from qadence.transpile import flatten, scale_primitive_blocks_only, transpile

        # make sure only leaves, i.e. primitive blocks are scaled
        block = transpile(flatten, scale_primitive_blocks_only)(observable)

        (native,) = convert_observable(block, n_qubits=n_qubits, config=self.config)
        return ConvertedObservable(native=native, abstract=block, original=observable)

    def assign_parameters(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor],
    ) -> Any:
        if param_values == {} and circuit.native.is_parametrized():
            missing = list(circuit.native.declared_variables.keys())
            raise ValueError(f"Please, provide values for the following parameters: {missing}")

        if param_values == {}:
            return circuit.native

        numpy_param_values = {
            k: v.detach().cpu().numpy()
            for (k, v) in param_values.items()
            if k in circuit.native.declared_variables
        }

        return circuit.native.build(**numpy_param_values)

    def _run(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        vals = to_list_of_dicts(param_values)

        # TODO: relax this constraint
        if self.config.cloud_configuration is not None:
            raise ValueError(
                "Cannot retrieve the wavefunction from cloud simulations. Do not"
                "specify any cloud credentials to use the .run() method"
            )

        state = state if state is None else _convert_init_state(state)
        batched_wf = np.zeros((len(vals), 2**circuit.abstract.n_qubits), dtype=np.complex128)

        for i, param_values_el in enumerate(vals):
            sequence = self.assign_parameters(circuit, param_values_el)
            pattern = circuit.original.register.device_specs.pattern
            if pattern is not None:
                add_addressing_pattern(sequence, pattern)
            sequence.measure()
            sim_result = simulate_sequence(sequence, self.config, state, n_shots=None)
            wf = (
                sim_result.get_final_state(  # type:ignore [union-attr]
                    ignore_global_phase=False, normalize=True
                )
                .full()
                .flatten()
            )
            # We flip the wavefunction coming out of pulser,
            # essentially changing logic 0 with logic 1 in the basis states.
            batched_wf[i] = np.flip(wf)

        batched_wf_torch = torch.from_numpy(batched_wf)

        if endianness != self.native_endianness:
            from qadence.transpile import invert_endianness

            batched_wf_torch = invert_endianness(batched_wf_torch)

        return batched_wf_torch

    def run_dm(
        self,
        circuit: ConvertedCircuit,
        noise: Noise,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list:
        vals = to_list_of_dicts(param_values)
        noise_probs = noise.options.get("noise_probs", None)
        if noise_probs is None:
            KeyError(f"A range of noise probabilies should be passed. Got {noise_probs}.")

        noisy_batched_dm = []

        # Pulser requires numpy types.
        for noise_prob in noise_probs.numpy():
            batched_dm = []
            sim_config = {"noise": noise.protocol, noise.protocol + "_prob": noise_prob}
            self.config.sim_config = SimConfig(**sim_config)

            for i, param_values_el in enumerate(vals):
                sequence = self.assign_parameters(circuit, param_values_el)
                sim_result = simulate_sequence(sequence, self.config, state)
                batched_dm.append(sim_result)
            noisy_batched_dm.append(batched_dm)

        return noisy_batched_dm

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        n_shots: int = 1,
        state: Tensor | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        if n_shots < 1:
            raise ValueError("You can only call sample with n_shots>0.")

        vals = to_list_of_dicts(param_values)
        state = state if state is None else _convert_init_state(state)

        samples = []
        for param_values_el in vals:
            sequence = self.assign_parameters(circuit, param_values_el)
            pattern = circuit.original.register.device_specs.pattern
            if pattern is not None:
                add_addressing_pattern(sequence, pattern)
            sequence.measure()
            sample = simulate_sequence(sequence, self.config, state, n_shots=n_shots)
            samples.append(sample)
        if endianness != self.native_endianness:
            from qadence.transpile import invert_endianness

            samples = invert_endianness(samples)
        if noise is not None:
            samples = apply_noise(noise=noise, samples=samples)
        if mitigation is not None:
            assert noise
            samples = apply_mitigation(noise=noise, mitigation=mitigation, samples=samples)
        return samples

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        observable = observable if isinstance(observable, list) else [observable]
        if mitigation is None:
            state = self.run(circuit, param_values=param_values, state=state, endianness=endianness)
            support = sorted(list(circuit.abstract.register.support))
            res_list = [
                obs.native(state, param_values, qubit_support=support) for obs in observable
            ]
            res = torch.transpose(torch.stack(res_list), 0, 1)
            res = res if len(res.shape) > 0 else res.reshape(1)
            return res.real
        elif mitigation is not None:
            mitigation_fn = mitigation.get_mitigation_fn()
            mitigated_exp_val = mitigation_fn(
                backend_name=self.name,
                circuit=circuit.original,
                observable=[obs.original for obs in observable],
                param_values=param_values,
                state=state,
                measurement=measurement,
                noise=noise,
                mitigation=mitigation,
                endianness=endianness,
            )
            return mitigated_exp_val

    @staticmethod
    def _overlap(bras: Tensor, kets: Tensor) -> Tensor:
        return overlap_exact(bras, kets)

    @staticmethod
    def default_configuration() -> Configuration:
        return Configuration()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass
from typing import Any, Callable

from qadence.backend import Backend, Converted, ConvertedCircuit, ConvertedObservable
from qadence.blocks import AbstractBlock, PrimitiveBlock
from qadence.blocks.utils import uuid_to_block
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.noise import Noise
from qadence.types import ArrayLike, DiffMode, Endianness, Engine, ParamDictType


@dataclass(frozen=True, eq=True)
class DifferentiableBackend(ABC):
    """The abstract class which wraps any (non)-natively differentiable QuantumBackend.

       in an automatic differentiation engine.

    Arguments:
        backend: An instance of the QuantumBackend type perform execution.
        engine: Which automatic differentiation engine the QuantumBackend runs on.
        diff_mode: A differentiable mode supported by the differentiation engine.
    """

    backend: Backend
    engine: Engine
    diff_mode: DiffMode

    # TODO: Add differentiable overlap calculation
    _overlap: Callable = None  # type: ignore [assignment]

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: ParamDictType = {},
        n_shots: int = 100,
        state: ArrayLike | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        """Sample bitstring from the registered circuit.

        Arguments:
            circuit: A backend native quantum circuit to be executed.
            param_values: The values of the parameters after embedding
            n_shots: The number of shots. Defaults to 1.
            state: Initial state.
            noise: A noise model to use.
            mitigation: A mitigation protocol to apply to noisy samples.
            endianness: Endianness of the resulting bitstrings.

        Returns:
            An iterable with all the sampled bitstrings
        """

        return self.backend.sample(
            circuit=circuit,
            param_values=param_values,
            n_shots=n_shots,
            state=state,
            noise=noise,
            mitigation=mitigation,
            endianness=endianness,
        )

    def run(
        self,
        circuit: ConvertedCircuit,
        param_values: ParamDictType = {},
        state: ArrayLike | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> ArrayLike:
        """Run on the underlying backend."""
        return self.backend.run(
            circuit=circuit, param_values=param_values, state=state, endianness=endianness
        )

    @abstractmethod
    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: ParamDictType = {},
        state: ArrayLike | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Any:
        """Compute the expectation value of the `circuit` with the given `observable`.

        Arguments:
            circuit: A converted circuit as returned by `backend.circuit`.
            observable: A converted observable as returned by `backend.observable`.
            param_values: _**Already embedded**_ parameters of the circuit. See
                [`embedding`][qadence.blocks.embedding.embedding] for more info.
            state: Initial state.
            measurement: Optional measurement protocol. If None, use
                exact expectation value with a statevector simulator.
            noise: A noise model to use.
            mitigation: The error mitigation to use.
            endianness: Endianness of the resulting bit strings.
        """
        raise NotImplementedError(
            "A DifferentiableBackend needs to override the expectation method."
        )

    def default_configuration(self) -> Any:
        return self.backend.default_configuration()

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        if self.diff_mode == DiffMode.GPSR:
            parametrized_blocks = list(uuid_to_block(circuit.block).values())
            non_prim_blocks = filter(
                lambda b: not isinstance(b, PrimitiveBlock), parametrized_blocks
            )
            if len(list(non_prim_blocks)) > 0:
                raise ValueError(
                    "The circuit contains non-primitive blocks that are currently\
                    not supported by the PSR differentiable mode."
                )
        return self.backend.circuit(circuit)

    def observable(self, observable: AbstractBlock, n_qubits: int) -> ConvertedObservable:
        if self.diff_mode != DiffMode.AD and observable is not None:
            msg = (
                f"Differentiation mode '{self.diff_mode}' does not support parametric observables."
            )
            if isinstance(observable, list):
                for obs in observable:
                    if obs.is_parametric:
                        raise ValueError(msg)
            else:
                if observable.is_parametric:
                    raise ValueError(msg)
        return self.backend.observable(observable, n_qubits)

    def convert(
        self,
        circuit: QuantumCircuit,
        observable: list[AbstractBlock] | AbstractBlock | None = None,
    ) -> Converted:
        return self.backend.convert(circuit, observable)

    def assign_parameters(self, circuit: ConvertedCircuit, param_values: ParamDictType) -> Any:
        return self.backend.assign_parameters(circuit, param_values)



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import Any

import jax
import jax.numpy as jnp
from horqrux.utils import zero_state
from jax.typing import ArrayLike

from qadence.backend import Backend as BackendInterface
from qadence.backend import ConvertedCircuit, ConvertedObservable
from qadence.backends.jax_utils import (
    tensor_to_jnp,
    unhorqify,
    uniform_batchsize,
)
from qadence.backends.utils import pyqify
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.noise import Noise
from qadence.transpile import flatten, scale_primitive_blocks_only, transpile
from qadence.types import BackendName, Endianness, Engine, ParamDictType
from qadence.utils import int_to_basis

from .config import Configuration, default_passes
from .convert_ops import HorqruxCircuit, convert_block, convert_observable


@dataclass(frozen=True, eq=True)
class Backend(BackendInterface):
    # set standard interface parameters
    name: BackendName = BackendName.HORQRUX  # type: ignore[assignment]
    supports_ad: bool = True
    supports_adjoint: bool = False
    support_bp: bool = True
    supports_native_psr: bool = False
    is_remote: bool = False
    with_measurements: bool = True
    with_noise: bool = False
    native_endianness: Endianness = Endianness.BIG
    config: Configuration = field(default_factory=Configuration)
    engine: Engine = Engine.JAX

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        passes = self.config.transpilation_passes
        if passes is None:
            passes = default_passes(self.config)

        original_circ = circuit
        if len(passes) > 0:
            circuit = transpile(*passes)(circuit)
        ops = convert_block(circuit.block, n_qubits=circuit.n_qubits, config=self.config)
        return ConvertedCircuit(
            native=HorqruxCircuit(ops), abstract=circuit, original=original_circ
        )

    def observable(self, observable: AbstractBlock, n_qubits: int) -> ConvertedObservable:
        transpilations = [
            flatten,
            scale_primitive_blocks_only,
        ]
        block = transpile(*transpilations)(observable)  # type: ignore[call-overload]
        hq_obs = convert_observable(block, n_qubits=n_qubits, config=self.config)
        return ConvertedObservable(native=hq_obs, abstract=block, original=observable)

    def _run(
        self,
        circuit: ConvertedCircuit,
        param_values: ParamDictType = {},
        state: ArrayLike | None = None,
        endianness: Endianness = Endianness.BIG,
        horqify_state: bool = True,
        unhorqify_state: bool = True,
    ) -> ArrayLike:
        """Propagate a state through a circuit given and retrieve the wave function.

        Args:
            circuit: A ConvertedCircuit object holding the native horqrux circuit.
            param_values: A dict holding the embedded parameters which the native ciruit expects.
            state: The input state.
            measurement: The protocol.
            horqify_state: Reshape the state into [2,2,...]
            unhorqify_state: Flatten the resulting state.

        Returns:
            A jax.Array of shape [2,2,...] if unhorqify=True else shape = [1, 2**n_qubits]
        """
        n_qubits = circuit.abstract.n_qubits
        if state is None:
            state = zero_state(n_qubits)
        else:
            state = tensor_to_jnp(pyqify(state)) if horqify_state else state
        state = circuit.native.forward(state, param_values)
        if endianness != self.native_endianness:
            state = jnp.reshape(state, (1, 2**n_qubits))  # batch_size is always 1
            ls = list(range(2**n_qubits))
            permute_ind = jnp.array([int(f"{num:0{n_qubits}b}"[::-1], 2) for num in ls])
            state = state[:, permute_ind]
        if unhorqify_state:
            state = unhorqify(state)
        return state

    def run_dm(
        self,
        circuit: ConvertedCircuit,
        noise: Noise,
        param_values: ParamDictType = {},
        state: ArrayLike | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> ArrayLike:
        raise NotImplementedError

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: ParamDictType = {},
        state: ArrayLike | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> ArrayLike:
        """Compute the expectation value of a circuit given a observable.

        Args:
            circuit: A ConvertedCircuit object holding the native horqrux circuit.
            observable: A ConvertedObservable object holding the native horqrux observable.
            param_values: A dict holding the embedded parameters which the native ciruit expects.
            state: The input state.
            measurement: The protocol.
            noise: The Noise.
            mitigation: Mitigation strategy.
            endianness (Endianness): The target endianness of the resulting samples.

        Returns:
            A jax.Array of shape (batch_size, n_observables)
        """
        observable = observable if isinstance(observable, list) else [observable]
        batch_size = max([arr.size for arr in param_values.values()])
        n_obs = len(observable)

        def _expectation(params: ParamDictType) -> ArrayLike:
            out_state = self.run(
                circuit, params, state, endianness, horqify_state=True, unhorqify_state=False
            )
            return jnp.array([o.native.forward(out_state, params) for o in observable])

        if batch_size > 1:  # We vmap for batch_size > 1
            expvals = jax.vmap(_expectation, in_axes=({k: 0 for k in param_values.keys()},))(
                uniform_batchsize(param_values)
            )
        else:
            expvals = _expectation(param_values)
        if expvals.size > 1:
            expvals = jnp.reshape(expvals, (batch_size, n_obs))
        else:
            expvals = jnp.squeeze(
                expvals, 0
            )  # For the case of batch_size == n_obs == 1, we remove the dims
        return expvals

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: ParamDictType = {},
        n_shots: int = 1,
        state: ArrayLike | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        """Samples from a batch of discrete probability distributions.

        Args:
            circuit: A ConvertedCircuit object holding the native horqrux circuit.
            param_values: A dict holding the embedded parameters which the native ciruit expects.
            n_shots: The number of samples to generate per distribution.
            state: The input state.
            noise: The Noise.
            mitigation: Mitigation strategy.
            endianness (Endianness): The target endianness of the resulting samples.

        Returns:
            A list of Counter objects where each key represents a bitstring
            and its value the number of times it has been sampled from the given wave function.
        """
        if n_shots < 1:
            raise ValueError("You can only call sample with n_shots>0.")

        def _sample(
            _probs: ArrayLike, n_shots: int, endianness: Endianness, n_qubits: int
        ) -> Counter:
            _logits = jax.vmap(lambda _p: jnp.log(_p / (1 - _p)))(_probs)

            def _smple(accumulator: ArrayLike, i: int) -> tuple[ArrayLike, None]:
                accumulator = accumulator.at[i].set(
                    jax.random.categorical(jax.random.PRNGKey(i), _logits)
                )
                return accumulator, None

            samples = jax.lax.scan(
                _smple, jnp.empty_like(jnp.arange(n_shots)), jnp.arange(n_shots)
            )[0]
            return Counter(
                {
                    int_to_basis(k=k, n_qubits=n_qubits, endianness=endianness): count.item()
                    for k, count in enumerate(jnp.bincount(samples))
                    if count > 0
                }
            )

        wf = self.run(
            circuit=circuit,
            param_values=param_values,
            state=state,
            horqify_state=True,
            unhorqify_state=False,
        )
        probs = jnp.abs(jnp.float_power(wf, 2.0)).ravel()
        samples = [
            _sample(
                _probs=probs,
                n_shots=n_shots,
                endianness=endianness,
                n_qubits=circuit.abstract.n_qubits,
            ),
        ]

        return samples

    def assign_parameters(self, circuit: ConvertedCircuit, param_values: ParamDictType) -> Any:
        raise NotImplementedError

    @staticmethod
    def _overlap(bras: ArrayLike, kets: ArrayLike) -> ArrayLike:
        # TODO
        raise NotImplementedError

    @staticmethod
    def default_configuration() -> Configuration:
        return Configuration()
